"""Evaluate statistical uncertainties from DL2 MC event files."""

import logging

import numpy as np
from astropy import units as u
from astropy.io import fits

__all__ = ["StatisticalErrorEvaluator"]


class StatisticalErrorEvaluator:
    """
    Evaluates statistical uncertainties from a DL2 MC event file.

    Parameters
    ----------
    file_path : str
        Path to the DL2 MC event file.
    file_type : str
        Type of the file, either 'point-like' or 'cone'.
    metrics : dict, optional
        Dictionary of metrics to evaluate. Default is None.
    grid_point : tuple, optional
        Tuple specifying the grid point (energy, azimuth, zenith, NSB, offset).
    """

    def __init__(
        self,
        file_path: str,
        file_type: str,
        metrics: dict[str, float],
        grid_point: tuple[float, float, float, float, float] | None = None,
    ):
        """Init the evaluator with a DL2 MC event file, its type, and metrics to calculate."""
        self._logger = logging.getLogger(__name__)
        self.file_type = file_type
        self.metrics = metrics
        self.grid_point = grid_point

        self.data = self.load_data_from_file(file_path)

        self.uncertainty_effective_area = None
        self.energy_estimate = None
        self.sigma_energy = None
        self.delta_energy = None

        self.metric_results = None
        self.energy_threshold = None

    def _load_event_data(self, hdul, data_type):
        """
        Load data and units for the event and simulated data data.

        Parameters
        ----------
        hdul : HDUList
            The HDUList object.
        data_type: str
            The type of data to load ('EVENTS' or 'SIMULATED EVENTS').

        Returns
        -------
        dict
            Dictionary containing units for the event data.
        """
        _data = hdul[data_type].data  # pylint: disable=E1101
        _header = hdul[data_type].header  # pylint: disable=E1101
        _units = {}
        for idx, col_name in enumerate(_data.columns.names, start=1):
            unit_key = f"TUNIT{idx}"
            if unit_key in _header:
                _units[col_name] = u.Unit(_header[unit_key])
            else:
                _units[col_name] = None
        return _data, _units

    def _set_grid_point(self, events_data):
        """Set azimuth/zenith angle of grid point."""
        unique_azimuths = np.unique(events_data["PNT_AZ"]) * u.deg
        unique_zeniths = 90 * u.deg - np.unique(events_data["PNT_ALT"]) * u.deg
        if len(unique_azimuths) > 1 or len(unique_zeniths) > 1:
            msg = (
                f"Multiple values found for azimuth ({unique_azimuths}) zenith ({unique_zeniths})."
            )
            self._logger.error(msg)
            raise ValueError(msg)
        if self.grid_point is not None:
            self._logger.warning(
                f"Grid point already set to: {self.grid_point}. "
                "Overwriting with new values from file."
            )
        self.grid_point = (
            1 * u.TeV,
            unique_azimuths[0],
            unique_zeniths[0],
            0,
            0 * u.deg,
        )
        self._logger.info(f"Grid point values: {self.grid_point}")

    def load_data_from_file(self, file_path):
        """
        Load data from the DL2 MC event file and return dictionaries with units.

        Returns
        -------
        dict
            Dictionary containing data from the DL2 MC event file with units.
        """
        data = {}
        try:
            with fits.open(file_path) as hdul:
                events_data, event_units = self._load_event_data(hdul, "EVENTS")
                sim_events_data, sim_units = self._load_event_data(hdul, "SIMULATED EVENTS")

                data = {
                    "event_energies_reco": events_data["ENERGY"] * event_units["ENERGY"],
                    "event_energies_mc": events_data["MC_ENERGY"] * event_units["MC_ENERGY"],
                    "bin_edges_low": sim_events_data["MC_ENERG_LO"] * sim_units["MC_ENERG_LO"],
                    "bin_edges_high": sim_events_data["MC_ENERG_HI"] * sim_units["MC_ENERG_HI"],
                    "simulated_event_histogram": sim_events_data["EVENTS"] * u.count,
                    "viewcone": hdul[3].data["viewcone"][0][1],  # pylint: disable=E1101
                    "core_range": hdul[3].data["core_range"][0][1],  # pylint: disable=E1101
                }
                self._set_grid_point(events_data)

        except FileNotFoundError as e:
            error_message = f"Error loading file {file_path}: {e}"
            self._logger.error(error_message)
            raise FileNotFoundError(error_message) from e
        return data

    def create_bin_edges(self):
        """
        Create unique energy bin edges.

        Returns
        -------
        bin_edges : array
            Array of unique energy bin edges.
        """
        bin_edges_low = self.data["bin_edges_low"]
        bin_edges_high = self.data["bin_edges_high"]
        bin_edges = np.concatenate([bin_edges_low, [bin_edges_high[-1]]])
        return np.unique(bin_edges)

    def compute_reconstructed_event_histogram(self, event_energies_reco, bin_edges):
        """
        Compute histogram of events as function of reconstructed energy.

        Parameters
        ----------
        event_energies_reco : array
            Array of reconstructed energy per event.
        bin_edges : array
            Array of energy bin edges.

        Returns
        -------
        reconstructed_event_histogram : array
            Histogram of reconstructed events.
        """
        event_energies_reco = event_energies_reco.to(bin_edges.unit)

        reconstructed_event_histogram, _ = np.histogram(
            event_energies_reco.value, bins=bin_edges.value
        )
        return reconstructed_event_histogram * u.count

    def compute_efficiency_and_errors(self, reconstructed_event_counts, simulated_event_counts):
        """
        Compute reconstructed event efficiency and its uncertainty assuming binomial distribution.

        Parameters
        ----------
        reconstructed_event_counts : array with units
            Histogram counts of reconstructed events.
        simulated_event_counts : array with units
            Histogram counts of simulated events.

        Returns
        -------
        efficiencies : array
            Array of calculated efficiencies.
        relative_errors : array
            Array of relative uncertainties.
        """
        # Ensure the inputs have compatible units
        reconstructed_event_counts = (
            reconstructed_event_counts.to(u.ct)
            if isinstance(reconstructed_event_counts, u.Quantity)
            else reconstructed_event_counts * u.ct
        )
        simulated_event_counts = (
            simulated_event_counts.to(u.ct)
            if isinstance(simulated_event_counts, u.Quantity)
            else simulated_event_counts * u.ct
        )

        if np.any(reconstructed_event_counts > simulated_event_counts):
            raise ValueError("Reconstructed event counts exceed simulated event counts.")

        # Compute efficiencies, ensuring the output is dimensionless
        efficiencies = np.divide(
            reconstructed_event_counts,
            simulated_event_counts,
            out=np.zeros_like(reconstructed_event_counts),
            where=simulated_event_counts > 0,
        ).to(u.dimensionless_unscaled)

        # Set up a mask for valid data with a unit-consistent threshold
        valid = (simulated_event_counts > 0) & (reconstructed_event_counts > 0)

        uncertainties = np.zeros_like(reconstructed_event_counts.value) * u.dimensionless_unscaled

        if np.any(valid):
            uncertainties[valid] = np.sqrt(
                np.maximum(
                    simulated_event_counts[valid]
                    / reconstructed_event_counts[valid]
                    * (1 - reconstructed_event_counts[valid] / simulated_event_counts[valid]),
                    0,
                )
            )

        # Compute relative errors
        relative_errors = np.divide(
            uncertainties,
            np.sqrt(simulated_event_counts.value),
            out=np.zeros_like(uncertainties, dtype=float),
            where=uncertainties > 0,
        )

        return efficiencies, relative_errors

    def calculate_energy_threshold(self, requested_eff_area_fraction=0.1):
        """
        Calculate the energy threshold where the effective area exceeds 10% of its maximum value.

        Returns
        -------
        float
            Energy threshold value.
        """
        bin_edges = self.create_bin_edges()
        reconstructed_event_histogram = self.compute_reconstructed_event_histogram(
            self.data["event_energies_mc"], bin_edges
        )
        simulated_event_histogram = self.data["simulated_event_histogram"]

        efficiencies, _ = self.compute_efficiency_and_errors(
            reconstructed_event_histogram, simulated_event_histogram
        )

        # Determine the effective area threshold (10% of max effective area)
        max_efficiency = np.max(efficiencies)
        threshold_efficiency = requested_eff_area_fraction * max_efficiency

        threshold_index = np.argmax(efficiencies >= threshold_efficiency)
        if threshold_index == 0 and efficiencies[0] < threshold_efficiency:
            return

        self.energy_threshold = bin_edges[threshold_index]

    def calculate_uncertainty_effective_area(self):
        """
        Calculate the uncertainties on the effective collection area.

        Returns
        -------
        errors : dict
            Dictionary with uncertainties for the file.
        """
        bin_edges = self.create_bin_edges()
        reconstructed_event_histogram = self.compute_reconstructed_event_histogram(
            self.data["event_energies_mc"], bin_edges
        )
        simulated_event_histogram = self.data["simulated_event_histogram"]
        _, relative_errors = self.compute_efficiency_and_errors(
            reconstructed_event_histogram, simulated_event_histogram
        )
        return {"relative_errors": relative_errors}

    def calculate_energy_estimate(self):
        """
        Calculate the uncertainties in energy estimation.

        Returns
        -------
        float
            The calculated uncertainty for energy estimation.
        """
        logging.info("Calculating Energy Resolution Error")

        event_energies_reco = self.data["event_energies_reco"]
        event_energies_mc = self.data["event_energies_mc"]

        if len(event_energies_reco) != len(event_energies_mc):
            raise ValueError(
                f"Mismatch in the number of energies: {len(event_energies_reco)} vs "
                f"{len(event_energies_mc)}"
            )

        energy_deviation = (event_energies_reco - event_energies_mc) / event_energies_mc

        bin_edges = self.create_bin_edges()
        bin_indices = np.digitize(event_energies_reco, bin_edges) - 1

        energy_deviation_by_bin = [
            energy_deviation[bin_indices == i] for i in range(len(bin_edges) - 1)
        ]

        # Calculate sigma for each bin
        sigma_energy = [np.std(d) if len(d) > 0 else np.nan for d in energy_deviation_by_bin]

        # Calculate delta_energy as the mean deviation for each bin
        delta_energy = [np.mean(d) if len(d) > 0 else np.nan for d in energy_deviation_by_bin]

        # Combine sigma into a single measure
        overall_uncertainty = np.nanmean(sigma_energy)

        return overall_uncertainty, sigma_energy, delta_energy

    def calculate_metrics(self):
        """Calculate all defined metrics as specified in self.metrics and store results."""
        if "uncertainty_effective_area" in self.metrics:
            self.uncertainty_effective_area = self.calculate_uncertainty_effective_area()
            if self.uncertainty_effective_area:
                energy_range = self.metrics.get("uncertainty_effective_area", {}).get(
                    "energy_range"
                )
                min_energy, max_energy = (
                    energy_range["value"][0] * u.Unit(energy_range["unit"]),
                    energy_range["value"][1] * u.Unit(energy_range["unit"]),
                )

                valid_errors = [
                    error
                    for energy, error in zip(
                        self.data["bin_edges_low"],
                        self.uncertainty_effective_area["relative_errors"],
                    )
                    if min_energy <= energy <= max_energy
                ]
                self.uncertainty_effective_area["max_error"] = (
                    max(valid_errors) if valid_errors else 0.0
                )
                ref_value = self.metrics.get("uncertainty_effective_area", {}).get("target_error")[
                    "value"
                ]
                self._logger.info(
                    f"Effective Area Error (max in validity range): "
                    f"{self.uncertainty_effective_area['max_error'].value:.6f}, "
                    f"Reference: {ref_value:.3f}"
                )

        if "energy_estimate" in self.metrics:
            self.energy_estimate, self.sigma_energy, self.delta_energy = (
                self.calculate_energy_estimate()
            )
            ref_value = self.metrics.get("energy_estimate", {}).get("target_error")["value"]
            self._logger.info(
                f"Energy Estimate Error: {self.energy_estimate:.3f}, Reference: {ref_value:.3f}"
            )
        else:
            raise ValueError("Invalid metric specified.")
        self.metric_results = {
            "uncertainty_effective_area": self.uncertainty_effective_area,
            "energy_estimate": self.energy_estimate,
        }
        return self.metric_results

    def calculate_max_error_for_effective_area(self):
        """
        Calculate the maximum relative error for effective area.

        Returns
        -------
        max_error : float
            Maximum relative error.
        """
        if "relative_errors" in self.metric_results["uncertainty_effective_area"]:
            return np.max(self.metric_results["uncertainty_effective_area"]["relative_errors"])
        if self.uncertainty_effective_area:
            return np.max(self.uncertainty_effective_area["relative_errors"])
        return None

    def calculate_overall_metric(self, metric="average"):
        """
        Calculate an overall metric for the statistical uncertainties.

        Parameters
        ----------
        metric : str
            The metric to calculate ('average', 'maximum').

        Returns
        -------
        dict
            Dictionary with overall maximum errors for each metric.
        """
        # Decide how to combine the metrics
        if self.metric_results is None:
            raise ValueError("Metrics have not been computed yet.")

        overall_max_errors = {}

        for metric_name, result in self.metric_results.items():
            if metric_name == "uncertainty_effective_area":
                max_errors = self.calculate_max_error_for_effective_area()
                overall_max_errors[metric_name] = max_errors if max_errors else 0
            elif metric_name in [
                "error_gamma_ray_psf",
            ]:
                overall_max_errors[metric_name] = result
            else:
                raise ValueError(f"Unsupported result type for {metric_name}: {type(result)}")
        self._logger.info(f"overall_max_errors {overall_max_errors}")
        all_max_errors = list(overall_max_errors.values())
        if metric == "average":
            overall_metric = np.mean(all_max_errors)
        elif metric == "maximum":
            overall_metric = np.max(all_max_errors)
        else:
            raise ValueError(f"Unsupported metric: {metric}")

        return overall_metric



#!/usr/bin/python3
"""Read model parameters and configuration from sim_telarray configuration files."""

import logging
import re

import astropy.units as u
import numpy as np

import simtools.utils.general as gen
from simtools.utils import names

__all__ = ["SimtelConfigReader"]


def get_list_of_simtel_parameters(simtel_config_file):
    """
    Return list of simtel parameters found in simtel configuration file.

    Parameters
    ----------
    simtel_config_file: str
        File name for sim_telarray configuration

    Returns
    -------
    list
        List of parameters found in simtel configuration file.

    """
    simtel_parameter_set = set()
    with open(simtel_config_file, encoding="utf-8") as file:
        for line in file:
            parts_of_lines = re.split(r",\s*|\s+", line.strip())
            simtel_parameter_set.add(parts_of_lines[1].lower())
    return sorted(simtel_parameter_set)


class SimtelConfigReader:
    """
    Reads model parameters from configuration files and converts to the simtools representation.

    The output format are simtool-db-style json dicts.
    Model parameters are read from sim_telarray configuration files.
    The sim_telarray configuration can be generated using e.g., the following sim_telarray command:

    ... code-block:: console

        sim_telarray/bin/sim_telarray \
            -c sim_telarray/cfg/CTA/CTA-PROD6-LaPalma.cfg\
            -C limits=no-internal -C initlist=no-internal -C list=no-internal\
            -C typelist=no-internal -C maximum_telescopes=30\
            -DNSB_AUTOSCALE -DNECTARCAM -DHYPER_LAYOUT\
            -DNUM_TELESCOPES=30 /dev/null 2>|/dev/null | grep '(@cfg)' | sed 's/^(@cfg)

    Parameters
    ----------
    schema_file: str
        Schema file describing the model parameter.
    simtel_config_file: str or Path
        Path of the file to read from.
    simtel_telescope_name: str
        Telescope name (sim_telarray convention)
    parameter_name: str
        Parameter name (default: read from schema file)
    camera_pixels: int
        Number of camera pixels
    """

    def __init__(
        self,
        schema_file,
        simtel_config_file,
        simtel_telescope_name,
        parameter_name=None,
        camera_pixels=None,
    ):
        """Initialize SimtelConfigReader."""
        self._logger = logging.getLogger(__name__)
        self._logger.debug("Init SimtelConfigReader")

        self.schema_file = schema_file
        self.schema_dict = (
            gen.collect_data_from_file(file_name=self.schema_file)
            if self.schema_file is not None
            else None
        )
        self.parameter_name = self.schema_dict.get("name") if self.schema_dict else parameter_name
        try:
            self.simtel_parameter_name = names.get_simulation_software_name_from_parameter_name(
                self.parameter_name
            ).upper()
        except (KeyError, AttributeError):
            self.simtel_parameter_name = self.parameter_name.upper()
        self.simtel_telescope_name = simtel_telescope_name
        self.camera_pixels = camera_pixels
        self.parameter_dict = self.read_simtel_config_file(
            simtel_config_file, simtel_telescope_name
        )

    def _should_skip_limits_check(self, data_type):
        """Check if limits should be skipped."""
        return data_type == "limits" and self.parameter_dict.get("type") == "bool"

    def _get_schema_values(self, data_type):
        """Check schema values for limits, unit, and default."""
        try:
            if data_type == "limits":
                _from_schema = [
                    self.schema_dict["data"][0]["allowed_range"].get("min"),
                    self.schema_dict["data"][0]["allowed_range"].get("max"),
                ]
                return _from_schema[0] if _from_schema[1] is None else _from_schema
            if len(self.schema_dict["data"]) == 1:
                return self.schema_dict["data"][0].get(data_type)
            return [data.get(data_type) for data in self.schema_dict["data"]]
        except (KeyError, IndexError):
            return None

    @staticmethod
    def _values_match(_from_simtel, _from_schema):
        """
        Check if values match (are close for floats).

        Convert where necessary astropy.Quantity to float.

        """
        if isinstance(_from_simtel, u.Quantity):
            _from_simtel = _from_simtel.value
        if isinstance(_from_simtel, np.ndarray) and len(_from_simtel) > 0:
            _from_simtel = np.array(
                [v.value if isinstance(v, u.Quantity) else v for v in _from_simtel]
            )
        try:
            if not isinstance(_from_schema, list | np.ndarray) and _from_simtel == _from_schema:
                return True
        except ValueError:
            pass

        try:
            if np.all(np.isclose(_from_simtel, _from_schema)):
                return True
        except (TypeError, ValueError):
            pass

        return False

    def _log_mismatch_warning(self, data_type, _from_simtel, _from_schema):
        """Log mismatch warning."""
        self._logger.warning(f"Values for {data_type} do not match:")
        self._logger.warning(
            f"  from simtel: {self.simtel_parameter_name} {_from_simtel} ({type(_from_simtel)})"
        )
        self._logger.warning(
            f"  from schema: {self.parameter_name} {_from_schema} ({type(_from_schema)})"
        )

    def compare_simtel_config_with_schema(self):
        """
        Compare limits and defaults reported by simtel_array with schema.

        This is mostly for debugging purposes and includes simple printing.
        Check for differences in 'default' and 'limits' entries.
        """
        for data_type in ["default", "limits"]:
            _from_simtel = self.parameter_dict.get(data_type)
            if self._should_skip_limits_check(data_type):
                continue

            _from_schema = self._get_schema_values(data_type)
            if isinstance(_from_schema, list):
                _from_schema = np.array(_from_schema, dtype=np.dtype(self.parameter_dict["type"]))

            if self._values_match(_from_simtel, _from_schema):
                self._logger.debug(f"Values for {data_type} match")
            else:
                self._log_mismatch_warning(data_type, _from_simtel, _from_schema)

    def read_simtel_config_file(self, simtel_config_file, simtel_telescope_name):
        """
        Read sim_telarray configuration file and return a dictionary with the parameter values.

        Parameters
        ----------
        simtel_config_file: str or Path
            Path of the file to read from.
        simtel_telescope_name: str
            Telescope name (sim_telarray convention)

        Returns
        -------
        dict
            Dictionary with the parameter values.

        """
        self._logger.debug(
            f"Reading simtel config file {simtel_config_file} for parameter {self.parameter_name}"
        )
        matching_lines = {}
        try:
            with open(simtel_config_file, encoding="utf-8") as file:
                for line in file:
                    # split line into parts (space, tabs, comma separated)
                    parts_of_lines = re.split(r",\s*|\s+", line.strip())
                    if self.simtel_parameter_name == parts_of_lines[1].upper():
                        matching_lines[parts_of_lines[0]] = parts_of_lines[2:]
        except FileNotFoundError as exc:
            self._logger.error(f"File {simtel_config_file} not found.")
            raise exc
        if len(matching_lines) == 0:
            self._logger.info(f"No entries found for parameter {self.simtel_parameter_name}")
            return None

        _para_dict = {}
        # first: extract line type (required for conversions and dimension)
        _para_dict["type"], _para_dict["dimension"] = self._get_type_and_dimension_from_simtel_cfg(
            matching_lines["type"]
        )
        # then: extract other fields
        # (order of keys matter; not all field are present for all parameters)
        for key in ["default", simtel_telescope_name, "limits"]:
            try:
                _para_dict[key], _ = self._add_value_from_simtel_cfg(
                    matching_lines[key],
                    dtype=_para_dict.get("type"),
                    n_dim=_para_dict.get("dimension"),
                    default=_para_dict.get("default"),
                    is_limit=(key == "limits"),
                )
            except KeyError:
                pass

        return _para_dict

    def _resolve_all_in_column(self, column):
        """
        Resolve 'all' entries in a column.

        This needs to resolve the following cases:
        no 'all' in any entry; ['all:', '5'], ['all: 5'], ['all:5', '3:1']
        This function is fine-tuned to the simtel configuration output.

        Parameters
        ----------
        column: list
            List of strings to resolve.

        Returns
        -------
        list
            List of resolved strings.

        """
        # don't do anything if all string items in column do not start with 'all'
        if not any(isinstance(item, str) and item.startswith("all") for item in column):
            return column, {}

        self._logger.debug(f"Resolving 'all' entries in column: {column}")
        # remove 'all:' entries
        column = [item for item in column if item not in ("all:", "all")]
        # resolve 'all:5' type entries
        column = [
            item.split(":")[1].replace(" ", "") if item.startswith("all:") else item
            for item in column
        ]
        # find 'index:value' type entries
        except_from_all = {}
        for item in column:
            if ":" in item:
                index, value = item.split(":")
                except_from_all[index] = value
        # finally remove entries containing ':'
        column = [item for item in column if ":" not in item]

        return column, except_from_all

    def _add_value_from_simtel_cfg(self, column, dtype=None, n_dim=1, default=None, is_limit=False):
        """
        Extract value(s) from simtel configuration file columns.

        This function is fine-tuned to the simtel configuration output.

        Parameters
        ----------
        column: list
            List of strings to extract value from.
        dtype: str
            Data type to convert value to.
        n_dim: int
            Length of array to be returned.
        default: object
            Default value to extend array to required length.

        Returns
        -------
        object, int
            Values extracted from column. Of object is a list of array, return length of array.

        """
        # string represents a lists of values (space or comma separated)
        if len(column) == 1:
            column = column[0].split(",") if "," in column[0] else column[0].split(" ")
        self._logger.debug(
            f"Adding value from simtel config: {column} (n_dim={n_dim}, default={default})"
        )
        column = [None if item.lower() == "none" else item for item in column]
        column, except_from_all = self._resolve_all_in_column(column)
        # extend array to required length (simtel uses sometimes 'all:' for all entries)
        if n_dim > 1 and len(column) < n_dim:
            try:
                # skip formatting: black reformats and violates E203
                column += default[len(column):]  # fmt: skip
            except TypeError:
                # extend array to required length using previous value
                column.extend([column[-1]] * (n_dim - len(column)))
        for index, value in except_from_all.items():
            column[int(index)] = value
        if dtype == "bool":
            column = np.array([bool(int(item)) for item in column])

        column, ndim = self._process_column(column, dtype)
        if not is_limit:
            column = self._add_units(column)
        return column, ndim

    def _add_units(self, column):
        """
        Add units as given in schema file to column.

        Take into account array types and dimensionless units.
        Ensure that integer values are returned as integers (astropy converts
        values to floats when multiplying them with units).

        """
        try:
            unit = self._get_schema_values("unit")
        except TypeError:  # no schema defined
            return column
        if unit is None or unit == "dimensionless":
            return column

        if isinstance(column, np.ndarray) and len(column) == len(unit):
            return np.array(
                [
                    col * u.Unit(un) if un != "dimensionless" else col
                    for col, un in zip(column, unit)
                ],
                dtype=object,
            )
        if isinstance(unit, str):
            column_with_unit = column * u.Unit(unit)
            if isinstance(column, int | np.integer):
                return u.Quantity(int(column_with_unit.value), unit, dtype=type(column))
            return column_with_unit

        return None

    def _process_column(self, column, dtype):
        """
        Process and return column prepared in _add_value_from_simtel_cfg.

        Parameters
        ----------
        column: list
            List of strings to process.
        dtype: str
            Data type to convert value to.
        """
        if len(column) == 1:
            if column[0] is not None:
                array_dtype = np.dtype(dtype) if dtype else None
                processed_value = np.array(column, dtype=array_dtype)[0]
                return processed_value, 1
            return None, 1
        if len(column) > 1:
            return np.array(column, dtype=np.dtype(dtype) if dtype else None), len(column)
        return None, None

    def _get_type_and_dimension_from_simtel_cfg(self, column):
        """
        Return type and dimension from simtel configuration column.

        'Func' type from simtel is treated as string. Return number
        of camera pixel for a hard-wired set up parameters.

        Parameters
        ----------
        column: list
            List of strings to extract value from.

        Returns
        -------
        str, int
            Type and dimension.

        """
        if column[0].lower() == "text" or column[0].lower() == "func":
            return "str", 1
        if column[0].lower() == "ibool":
            return "bool", int(column[1])
        if self.camera_pixels is not None and self.simtel_parameter_name in ["NIGHTSKY_BACKGROUND"]:
            return str(np.dtype(column[0].lower())), self.camera_pixels
        return str(np.dtype(column[0].lower())), int(column[1])



"""Reads the content of either a single histogram or simtel_array output file."""

import copy
import logging
from pathlib import Path

import numpy as np
from astropy import units as u
from astropy.table import QTable
from ctao_cr_spectra.definitions import IRFDOC_PROTON_SPECTRUM
from ctao_cr_spectra.spectral import cone_solid_angle
from eventio import EventIOFile, Histograms
from eventio.search_utils import yield_toplevel_of_type
from eventio.simtel import MCRunHeader

__all__ = [
    "HistogramIdNotFoundError",
    "InconsistentHistogramFormatError",
    "SimtelIOHistogram",
]


class InconsistentHistogramFormatError(Exception):
    """Exception for bad histogram format."""


class HistogramIdNotFoundError(Exception):
    """Exception for histogram ID not found."""


class SimtelIOHistogram:
    """
    Reads and generates histograms from sim_telarray output.

    Read the content of either a single histogram (.hdata, or .hdata.zst) or a single simtel_array
    output file (.simtel or .simtel.zst).

    Parameters
    ----------
    histogram_file: str
        The histogram (.hdata.zst) or simtel_array (.simtel.zst) file.
    area_from_distribution: bool
        If true, the area thrown (the area in which the simulated events are distributed)
        in the trigger rate calculation is estimated based on the event distribution.
        The expected shape of the distribution of events as function of the core distance is
        triangular up to the maximum distance. The weighted mean radius of the triangular
        distribution is 2/3 times the upper edge. Therefore, when using the
        ``area_from_distribution`` flag, the mean distance times 3/2, returns just the position of
        the upper edge in the triangle distribution with little impact of the binning and little
        dependence on the scatter area defined in the simulation. This is special useful when
        calculating trigger rate for individual telescopes.
        If false, the area thrown is estimated based on the maximum distance as given in
        the simulation configuration.
    energy_range: list
        The energy range used in the simulation. It must be passed as a list of floats and the
        energy must be in TeV.
        This argument is only needed and used if histogram_file is a .hdata file, in which case the
        energy range cannot be retrieved directly from the file.
    view_cone: list
        The view cone used in the simulation. It must be passed as a list of floats and the
        view cone must be in deg.
        This argument is only needed and used if histogram_file is a .hdata file, in which case the
        view cone cannot be retrieved directly from the file.
    """

    def __init__(
        self, histogram_file, area_from_distribution=False, energy_range=None, view_cone=None
    ):
        """Initialize SimtelIOHistogram class."""
        self._logger = logging.getLogger(__name__)
        self.histogram_file = histogram_file
        if not Path(histogram_file).exists():
            msg = f"File {histogram_file} does not exist."
            self._logger.error(msg)
            raise FileNotFoundError

        self._config = None
        self._total_area = None
        self._solid_angle = None
        self._histogram = None
        self._total_num_simulated_events = None
        self._total_num_triggered_events = None
        self._initialize_histogram()
        self.trigger_rate = None
        self.trigger_rate_uncertainty = None
        self.trigger_rate_per_energy_bin = None
        self.energy_axis = None
        self.radius_axis = None
        self.area_from_distribution = area_from_distribution

        self._set_view_cone(view_cone)
        self._set_energy_range(energy_range)

    def _initialize_histogram(self):
        """
        Initialize lists of histograms and files.

        Returns
        -------
        list:
            List of histograms.
        """
        with EventIOFile(self.histogram_file) as f:
            for obj in yield_toplevel_of_type(f, Histograms):
                self.histogram = obj.parse()

    @property
    def number_of_histogram_types(self):
        """Return number of histograms."""
        return len(self.histogram)

    def get_histogram_type_title(self, histogram_index):
        """
        Return the title of the histogram with index histogram_index.

        Parameters
        ----------
        histogram_index: int
            Histogram index.

        Returns
        -------
        str
            Histogram title.
        """
        return self.histogram[histogram_index]["title"]

    @property
    def config(self):
        """
        Return information about the input parameters for the simulation.

        If the file is a .hdata or .hdata.zst, config will be None.

        Returns
        -------
        dict:
            dictionary with information about the simulation (pyeventio MCRunHeader object).
        """
        if self._config is None:
            with EventIOFile(self.histogram_file) as f:
                self._config = next(
                    (obj.parse() for obj in f if isinstance(obj, MCRunHeader)), None
                )
        return self._config

    @property
    def total_number_of_events(self):
        """
        Return the total number of simulated and triggered events in the histograms.

        Returns
        -------
        int, int:
            total number of simulated and triggered events.
        """
        if self._total_num_simulated_events is None or self._total_num_triggered_events is None:
            simulated_histogram, trigger_histogram = self.fill_event_histogram_dicts()
            self._total_num_simulated_events = np.sum(simulated_histogram["data"])
            self._total_num_triggered_events = np.sum(trigger_histogram["data"])
            logging.debug(
                "Number of total simulated / triggered events: "
                f"{self._total_num_simulated_events} / {self._total_num_triggered_events}"
            )
        return self._total_num_simulated_events, self._total_num_triggered_events

    def fill_event_histogram_dicts(self):
        """
        Get data from the total simulated event and the triggered event histograms.

        Returns
        -------
        tuple(dict, dict):
            Information about the histograms with simulated and triggered events.

        Raises
        ------
        HistogramIdNotFoundError:
            if histogram ids not found. Problem with the file.
        """
        histograms = {hist["id"]: hist for hist in self.histogram if hist["id"] in {1, 2}}
        if 1 in histograms and 2 in histograms:
            return histograms[1], histograms[2]
        self._logger.error("Histograms ids not found. Please check your files.")
        raise HistogramIdNotFoundError

    def _set_view_cone(self, view_cone):
        """
        View cone used in the simulation.

        Parameters
        ----------
        view_cone: list
        The view cone used in the simulation. It must be passed as a list of floats and the
        view cone must be in deg (as in the CORSIKA configuration).

        Raises
        ------
        ValueError:
            if input parameter is missing.
        """
        if view_cone is None:
            try:
                self.view_cone = self.config["viewcone"] * u.deg
            except TypeError as exc:
                msg = (
                    "view_cone needs to be passed as argument (minimum and maximum of the "
                    "view cone radius in deg)."
                )
                self._logger.error(msg)
                raise ValueError(msg) from exc
        else:
            if isinstance(view_cone, u.Quantity):
                self.view_cone = view_cone.to(u.deg)
            else:
                self.view_cone = view_cone * u.deg

    @property
    def solid_angle(self):
        """
        Solid angle corresponding to the view cone.

        Returns
        -------
        astropy.Quantity[u.sr]:
            Solid angle corresponding to the view cone.
        """
        if self._solid_angle is None:
            self._solid_angle = cone_solid_angle(self.view_cone[1]) - cone_solid_angle(
                self.view_cone[0]
            )
        return self._solid_angle

    @property
    def total_area(self):
        """
        Total area covered by the simulated events (original CORSIKA CSCAT), i.e., area thrown.

        Returns
        -------
        astropy.Quantity[area]:
            Total area covered on the ground covered by the simulation.
        """
        if self._total_area is None:
            events_histogram, _ = self.fill_event_histogram_dicts()
            self._initialize_histogram_axes(events_histogram)

            if self.area_from_distribution is True:
                area_from_distribution_max_radius = 1.5 * np.average(
                    self.radius_axis[:-1], weights=np.sum(events_histogram["data"], axis=0)
                )
                self._total_area = (np.pi * (area_from_distribution_max_radius * u.m) ** 2).to(
                    u.cm**2
                )
            else:
                # The max of the core range is always half the upper edge:
                # self.radius_axis[-1]/2 is equal to self.config["core_range"][1]
                self._total_area = (
                    np.pi * (((self.radius_axis[-1] / 2 - self.radius_axis[0]) * u.m).to(u.cm)) ** 2
                )
        return self._total_area

    def _set_energy_range(self, energy_range):
        """
        Set energy range to be used in the simulations.

        Parameters
        ----------
        energy_range: list
        The energy range used in the simulation. It must be passed as a list of floats and the
        energy must be in TeV.

        Raises
        ------
        ValueError:
            if input parameter is missing.
        """
        if energy_range is None:
            try:
                self.energy_range = [
                    self.config["E_range"][0] * u.TeV,
                    self.config["E_range"][1] * u.TeV,
                ]
            except TypeError as exc:  # E_range not in self.config
                msg = (
                    "energy_range needs to be passed as argument (minimum and maximum"
                    " energies in TeV)."
                )
                self._logger.error(msg)
                raise ValueError(msg) from exc
        else:
            if isinstance(energy_range, u.Quantity):
                self.energy_range = energy_range.to(u.TeV)
            else:
                self.energy_range = energy_range * u.TeV

    @staticmethod
    def _produce_triggered_to_sim_fraction_hist(events_histogram, triggered_events_histogram):
        """
        Produce a new histogram with the fraction of triggered events over the simulated events.

        The dimension of the histogram is reduced, as the rates are summed for all the bins in
        impact distance.

        Parameters
        ----------
        events_histogram:
            A dictionary with "data" corresponding to a 2D histogram (impact distance x energy)
            for the simulated events.
        triggered_events_histogram:
            A dictionary with "data" corresponding to a 2D histogram (impact distance x energy)
            for the triggered events.

        Returns
        -------
        event_ratio_histogram:
            The new histogram with the fraction of triggered over simulated events.
        """
        simulated_events_per_energy_bin = np.sum(events_histogram["data"], axis=1)

        triggered_events_per_energy_bin = np.sum(triggered_events_histogram["data"], axis=1)
        ratio_per_energy_bin = np.zeros_like(triggered_events_per_energy_bin, dtype=float)

        non_zero_indices = np.nonzero(simulated_events_per_energy_bin)[0]
        ratio_per_energy_bin[non_zero_indices] = (
            triggered_events_per_energy_bin[non_zero_indices]
            / simulated_events_per_energy_bin[non_zero_indices]
        )
        return ratio_per_energy_bin

    def compute_system_trigger_rate(self, events_histogram=None, triggered_events_histogram=None):
        """
        Compute the system trigger rate and its uncertainty, which are saved as class attributes.

        If events_histogram and triggered_events_histogram are passed, they are used to calculate
        the trigger rate and trigger rate uncertainty, instead of the histograms from the file.
        This is specially useful when calculating the trigger rate for stacked files, in which case
        one can pass the histograms resulted from stacking the files to this function.
        Default is filling from the file.

        Parameters
        ----------
        events_histogram:
            A dictionary with "data" corresponding to a 2D histogram (core distance x energy)
            for the simulated events.
        triggered_events_histogram:
            A dictionary with "data" corresponding to a 2D histogram (core distance x energy)
            for the triggered events.
        """
        if self.trigger_rate is None:
            # Get the simulated and triggered 2D histograms from the simtel_array output file
            if events_histogram is None and triggered_events_histogram is None:
                events_histogram, triggered_events_histogram = self.fill_event_histogram_dicts()
            # Calculate triggered/simulated event 1D histogram (energy dependent)
            triggered_to_sim_fraction_hist = self._produce_triggered_to_sim_fraction_hist(
                events_histogram, triggered_events_histogram
            )
            self._initialize_histogram_axes(triggered_events_histogram)

            # Getting the particle distribution function according to the reference
            particle_distribution_function = self.get_particle_distribution_function(
                label="reference"
            )

            # Integrating the flux between the consecutive energy bins. The result given in
            # cm-2s-1sr-1
            flux_per_energy_bin = self._integrate_in_energy_bin(
                particle_distribution_function, self.energy_axis
            )

            # Derive the trigger rate per energy bin
            self.trigger_rate_per_energy_bin = (
                triggered_to_sim_fraction_hist
                * flux_per_energy_bin
                * self.total_area
                * self.solid_angle
            ).decompose()

            # Derive the system trigger rate
            self.trigger_rate = np.sum(self.trigger_rate_per_energy_bin)

            # Derive the uncertainty in the system trigger rate estimate
            self.trigger_rate_uncertainty = self.estimate_trigger_rate_uncertainty(
                self.trigger_rate,
                np.sum(events_histogram["data"]),
                np.sum(triggered_events_histogram["data"]),
            )

        return self.trigger_rate, self.trigger_rate_uncertainty

    def trigger_info_in_table(self):
        """
        Provide the trigger rate per energy bin in tabulated form.

        Returns
        -------
        astropy.QTable:
            The QTable instance with the trigger rate per energy bin.
        """
        meta = self.produce_trigger_meta_data()
        return QTable(
            [self.energy_axis[:-1] * u.TeV, (self.trigger_rate_per_energy_bin.to(u.Hz))],
            names=("Energy (TeV)", "Trigger rate (Hz)"),
            meta=meta,
        )

    def produce_trigger_meta_data(self):
        """
        Produce the meta data to include in the tabulated form of the trigger rate per energy bin.

        It shows some information from the input file (simtel_array file) and the final estimate
        system trigger rate.

        Returns
        -------
        dict:
            dictionary with the metadata.
        """
        return {
            "simtel_array_file": self.histogram_file,
            "simulation_input": self.print_info(mode="silent"),
            "system_trigger_rate (Hz)": self.trigger_rate.value,
        }

    def _integrate_in_energy_bin(self, particle_distribution_function, energy_axis):
        """
        Integrate the particle distribution.

        The function integrates between the consecutive energy bins given by the energy_axis array.

        Parameters
        ----------
        particle_distribution_function: ctao_cr_spectra.spectral.PowerLaw
            The function describing the spectral distribution.
        energy_axis: numpy.array
            The array with the simulated particle energies.

        Returns
        -------
        astropy.Quantity:
            astropy.Quantity of a numpy array with the energy integrated flux.
        """
        unit = None
        flux_per_energy_bin = []
        for i_energy, _ in enumerate(energy_axis[:-1]):
            integrated_flux = particle_distribution_function.integrate_energy(
                energy_axis[i_energy] * u.TeV, energy_axis[i_energy + 1] * u.TeV
            ).decompose(bases=[u.s, u.cm, u.sr])
            if unit is None:
                unit = integrated_flux.unit

            flux_per_energy_bin.append(integrated_flux.value)

        return np.array(flux_per_energy_bin) * unit

    def _initialize_histogram_axes(self, events_histogram):
        """
        Initialize the two axes of a histogram.

        The two axes are: the array with the edges of the bins in core
        distance and the edges of the array with the energy bins.

        Parameters
        ----------
        events_histogram:
            A single histogram from where to extract axis information.
        """
        self.radius_axis = np.linspace(
            events_histogram["lower_x"],
            events_histogram["upper_x"],
            events_histogram["n_bins_x"] + 1,
            endpoint=True,
        )

        self.energy_axis = np.logspace(
            events_histogram["lower_y"],
            events_histogram["upper_y"],
            events_histogram["n_bins_y"] + 1,
            endpoint=True,
        )

    def get_particle_distribution_function(self, label="reference"):
        """
        Get the particle distribution function.

        This depends on whether one wants the reference CR distribution or the distribution
        used in the simulation. This is controlled by label.
        By using label="reference", one gets the distribution function according to a pre-defined CR
        distribution, while by using label="simulation", the spectral index of the distribution
        function from the simulation is used. The label="simulation" works only when the
        input file is a .simtel file and not a .hdata file.

        Parameters
        ----------
        label: str
            label defining which distribution function. Possible values are: "reference", or
            "simulation".

        Returns
        -------
        ctao_cr_spectra.spectral.PowerLaw
            The function describing the spectral distribution.
        """
        if label == "reference":
            particle_distribution_function = copy.copy(IRFDOC_PROTON_SPECTRUM)
        elif label == "simulation":
            particle_distribution_function = self._get_simulation_spectral_distribution_function()
        else:
            msg = f"label {label} is not valid. Please use either 'reference' or 'simulation'."
            self._logger.error(msg)
            raise ValueError(msg)
        return particle_distribution_function

    def _get_simulation_spectral_distribution_function(self):
        """
        Get the simulation particle energy distribution according to its configuration.

        Returns
        -------
        ctao_cr_spectra.spectral.PowerLaw
            The function describing the spectral distribution.

        Raises
        ------
        ValueError:
            if input parameter is missing.
        """
        spectral_distribution = copy.copy(IRFDOC_PROTON_SPECTRUM)
        try:
            spectral_distribution.index = self.config["spectral_index"]
        except TypeError as exc:
            msg = (
                "spectral_index not found in the configuration of the file. "
                "Consider using a .simtel file instead."
            )
            self._logger.error(msg)
            raise ValueError(msg) from exc
        return spectral_distribution

    def estimate_observation_time(self, stacked_num_simulated_events=None):
        """
        Estimates the observation time corresponding to the simulated number of events.

        It uses the CTAO reference cosmic-ray spectra, the total number of particles simulated,
        and other information from the simulation configuration self.config.
        If stacked_num_simulated_events is given, the observation time is estimated from it instead
        of from the simulation configuration (useful for the stacked trigger rate estimate).

        Parameters
        ----------
        stacked_num_simulated_events: int
            total number of simulated events for the stacked dataset.

        Returns
        -------
        astropy.Quantity[time]
            Estimated observation time based on the total number of particles simulated.
        """
        first_estimate = IRFDOC_PROTON_SPECTRUM.compute_number_events(
            self.view_cone[0],
            self.view_cone[1],
            1 * u.s,
            self.total_area,
            self.energy_range[0],
            self.energy_range[1],
        )
        if stacked_num_simulated_events is None:
            _simulated_events, _ = self.total_number_of_events
            return (_simulated_events / first_estimate) * u.s
        return (stacked_num_simulated_events / first_estimate) * u.s

    def estimate_trigger_rate_uncertainty(
        self, trigger_rate_estimate, num_simulated_events, num_triggered_events
    ):
        """
        Estimate the trigger rate uncertainty.

        The calculation is based on the number of simulated and triggered events.
        Poisson statistics are assumed. The uncertainty is calculated based on propagation
        of the individual uncertainties.
        If stacked_num_simulated_events is passed, the uncertainty is estimated based on it instead
        of based on the total number of trigger events from the simulation configuration
        (useful for the stacked trigger rate estimate).

        Parameters
        ----------
        trigger_rate_estimate: astropy.Quantity[1/time]
            The already estimated the trigger rate.
        num_simulated_events: int
            Total number of simulated events.
        num_triggered_events: int
            Total number of triggered events.

        Returns
        -------
        astropy.Quantity[1/time]
            Uncertainty in the trigger rate estimate.
        """
        # pylint: disable=E1101
        return (
            trigger_rate_estimate.value
            * np.sqrt(1 / num_triggered_events + 1 / num_simulated_events)
        ) * trigger_rate_estimate.unit

    def print_info(self, mode=None):
        """
        Print information on the geometry and input parameters.

        Returns
        -------
        dict:
            Dictionary with the information, e.g., view angle, energy range, etc.
        """
        _simulated, _triggered = self.total_number_of_events
        info_dict = {
            "view_cone": self.view_cone,
            "solid_angle": self.solid_angle,
            "total_area": self.total_area,
            "energy_range": self.energy_range,
            "total_num_simulated_events": _simulated,
            "total_num_triggered_events": _triggered,
        }
        if mode != "silent":
            print(info_dict)
        return info_dict



"""Definition and modeling of a calibration device."""

import logging

from simtools.model.model_parameter import ModelParameter

__all__ = ["CalibrationModel"]


class CalibrationModel(ModelParameter):
    """
    CalibrationModel represents the MC model of an individual calibration device.

    It provides functionality to read the required parameters from the DB.

    Parameters
    ----------
    site: str
        Site name (e.g., South or North).
    calibration_device_model_name: str
        Calibration device model name (ex. ILLS-01, ILLN-01, ...).
    mongo_db_config: dict
        MongoDB configuration.
    model_version: str
        Model version.
    label: str, optional
        Instance label. Important for output file naming.
    """

    def __init__(
        self,
        site: str,
        calibration_device_model_name: str,
        mongo_db_config: dict,
        model_version: str,
        label: str | None = None,
    ):
        """Initialize CalibrationModel."""
        super().__init__(
            site=site,
            array_element_name=calibration_device_model_name,
            collection="calibration_devices",
            mongo_db_config=mongo_db_config,
            model_version=model_version,
            db=None,
            label=label,
        )

        self._logger = logging.getLogger(__name__)
        self._logger.debug("Init CalibrationModel %s %s", site, calibration_device_model_name)



#!/usr/bin/python3

"""
Make a regular array of telescopes and save it as astropy table.

The arrays consist of one telescope at the center of the array and or of 4 telescopes
in a square grid. These arrays are used for trigger rate simulations.

The array layout files created will be available at the data/layout directory.

Command line arguments
----------------------
site (str, required)
    observatory site (e.g., North or South).
model_version (str, optional)
    Model version to use (e.g., 6.0.0). If not provided, the latest version is used.

Example
-------
Runtime < 10 s.

.. code-block:: console

    simtools-generate-regular-arrays --site=North
"""

import logging
from pathlib import Path

import astropy.units as u
from astropy.table import QTable

import simtools.data_model.model_data_writer as writer
import simtools.utils.general as gen
from simtools.configuration import configurator
from simtools.utils import names

# Telescope distances for 4 tel square arrays
# !HARDCODED
telescope_distance = {"LST": 57.5 * u.m, "MST": 70 * u.m, "SST": 80 * u.m}


def _parse():
    config = configurator.Configurator(
        label=Path(__file__).stem,
        description=(
            "Generate a regular array of telescope and save as astropy table.\n"
            "Default telescope distances for 4 telescope square arrays are: \n"
            f"  LST: {telescope_distance['LST']}\n"
            f"  MST: {telescope_distance['MST']}\n"
            f"  SST: {telescope_distance['SST']}\n"
        ),
    )
    return config.initialize(
        db_config=False, simulation_model=["site", "model_version"], output=True
    )


def main():
    """Create layout array files (ecsv) of regular arrays."""
    args_dict, _ = _parse()

    logger = logging.getLogger()
    logger.setLevel(gen.get_log_level_from_user(args_dict["log_level"]))

    if args_dict["site"] == "South":
        array_list = ["1SST", "4SST", "1MST", "4MST", "1LST", "4LST"]
    else:
        array_list = ["1MST", "4MST", "1LST", "4LST"]

    for array_name in array_list:
        logger.info(f"Processing array {array_name}")

        tel_name, pos_x, pos_y, pos_z = [], [], [], []
        tel_size = array_name[1:4]

        # Single telescope at the center
        if array_name[0] == "1":
            tel_name.append(
                names.generate_array_element_name_from_type_site_id(
                    tel_size, args_dict["site"], "01"
                )
            )
            pos_x.append(0 * u.m)
            pos_y.append(0 * u.m)
            pos_z.append(0 * u.m)
        # 4 telescopes in a regular square grid
        else:
            for i in range(1, 5):
                tel_name.append(
                    names.generate_array_element_name_from_type_site_id(
                        tel_size, args_dict["site"], f"0{i}"
                    )
                )
                pos_x.append(telescope_distance[tel_size] * (-1) ** (i // 2))
                pos_y.append(telescope_distance[tel_size] * (-1) ** (i % 2))
                pos_z.append(0 * u.m)

        table = QTable(meta={"array_name": array_name, "site": args_dict["site"]})
        table["telescope_name"] = tel_name
        table["position_x"] = pos_x
        table["position_y"] = pos_y
        table["position_z"] = pos_z
        table.sort("telescope_name")
        table.pprint()

        output_file = args_dict.get("output_file")
        if output_file:
            output_path = Path(output_file)
            output_file = output_path.with_name(
                f"{output_path.stem}-{args_dict['site']}-{array_name}{output_path.suffix}"
            )
        writer.ModelDataWriter.dump(
            args_dict=args_dict,
            output_file=output_file,
            metadata=None,
            product_data=table,
        )


if __name__ == "__main__":
    main()



"""Module providing functionality to read and validate dictionaries using schema."""

import logging
from pathlib import Path

import jsonschema

import simtools.utils.general as gen
from simtools.constants import (
    METADATA_JSON_SCHEMA,
    MODEL_PARAMETER_METASCHEMA,
    MODEL_PARAMETER_SCHEMA_PATH,
    SCHEMA_PATH,
)
from simtools.data_model import format_checkers
from simtools.utils import names

_logger = logging.getLogger(__name__)


def get_get_model_parameter_schema_files(schema_directory=MODEL_PARAMETER_SCHEMA_PATH):
    """
    Return list of parameters and schema files located in schema file directory.

    Returns
    -------
    list
        List of parameters found in schema file directory.
    list
        List of schema files found in schema file directory.

    """
    schema_files = sorted(Path(schema_directory).rglob("*.schema.yml"))
    if not schema_files:
        raise FileNotFoundError(f"No schema files found in {schema_directory}")
    parameters = []
    for schema_file in schema_files:
        schema_dict = gen.collect_data_from_file(file_name=schema_file)
        parameters.append(schema_dict.get("name"))
    return parameters, schema_files


def get_model_parameter_schema_file(parameter):
    """
    Return schema file path for a given model parameter.

    Parameters
    ----------
    parameter: str
        Model parameter name.

    Returns
    -------
    Path
        Schema file path.

    """
    schema_file = MODEL_PARAMETER_SCHEMA_PATH / f"{parameter}.schema.yml"
    if not schema_file.exists():
        raise FileNotFoundError(f"Schema file not found: {schema_file}")
    return schema_file


def get_model_parameter_schema_version(schema_version=None):
    """
    Validate  and return schema versions.

    If no schema_version is given, the most recent version is provided.

    Parameters
    ----------
    schema_version: str
        Schema version.

    Returns
    -------
    str
        Schema version.

    """
    schemas = gen.collect_data_from_file(MODEL_PARAMETER_METASCHEMA)

    if schema_version is None and schemas:
        return schemas[0].get("version")

    if any(schema.get("version") == schema_version for schema in schemas):
        return schema_version

    raise ValueError(f"Schema version {schema_version} not found in {MODEL_PARAMETER_METASCHEMA}.")


def validate_dict_using_schema(data, schema_file=None, json_schema=None):
    """
    Validate a data dictionary against a schema.

    Parameters
    ----------
    data
        dictionary to be validated
    schema_file (dict)
        schema used for validation

    Raises
    ------
    jsonschema.exceptions.ValidationError
        if validation fails

    """
    if json_schema is None and schema_file is None:
        _logger.warning(f"No schema provided for validation of {data}")
        return None
    if json_schema is None:
        json_schema = load_schema(
            schema_file,
            data.get("schema_version")
            or data.get(
                "SCHEMA_VERSION", "0.1.0"
            ),  # default version to ensure backward compatibility
        )

    try:
        jsonschema.validate(data, schema=json_schema, format_checker=format_checkers.format_checker)
    except jsonschema.exceptions.ValidationError as exc:
        _logger.error(f"Validation failed using schema: {json_schema} for data: {data}")
        raise exc
    if (
        isinstance(data, dict)
        and data.get("meta_schema_url")
        and not gen.url_exists(data["meta_schema_url"])
    ):
        raise FileNotFoundError(f"Meta schema URL does not exist: {data['meta_schema_url']}")

    _logger.debug(f"Successful validation of data using schema ({json_schema.get('name')})")
    return data


def load_schema(schema_file=None, schema_version=None):
    """
    Load parameter schema from file.

    Parameters
    ----------
    schema_file: str
        Path to schema file.
    schema_version: str
        Schema version.

    Returns
    -------
    schema: dict
        Schema dictionary.

    Raises
    ------
    FileNotFoundError
        if schema file is not found

    """
    schema_file = schema_file or METADATA_JSON_SCHEMA

    for path in (schema_file, SCHEMA_PATH / schema_file):
        try:
            schema = gen.collect_data_from_file(file_name=path)
            break
        except FileNotFoundError:
            continue
    else:
        raise FileNotFoundError(f"Schema file not found: {schema_file}")

    if isinstance(schema, list):  # schema file with several schemas defined
        if schema_version is None:
            raise ValueError(f"Schema version not given in {schema_file}.")
        schema = next((doc for doc in schema if doc.get("version") == schema_version), None)
        if schema is None:
            raise ValueError(f"Schema version {schema_version} not found in {schema_file}.")
    elif schema_version is not None and schema_version != schema.get("version"):
        _logger.warning(f"Schema version {schema_version} does not match {schema.get('version')}")

    _logger.debug(f"Loading schema from {schema_file}")
    _add_array_elements("InstrumentTypeElement", schema)

    return schema


def _add_array_elements(key, schema):
    """
    Add list of array elements to schema.

    Avoids having to list all array elements in multiple schema.
    Assumes an element [key]['enum'] is a list of elements.

    Parameters
    ----------
    key: str
        Key in schema dictionary
    schema: dict
        Schema dictionary

    Returns
    -------
    dict
        Schema dictionary with added array elements.

    """
    _list_of_array_elements = sorted(names.array_elements().keys())

    def recursive_search(sub_schema, key):
        if key in sub_schema:
            if "enum" in sub_schema[key] and isinstance(sub_schema[key]["enum"], list):
                sub_schema[key]["enum"] = list(
                    set(sub_schema[key]["enum"] + _list_of_array_elements)
                )
            else:
                sub_schema[key]["enum"] = _list_of_array_elements
        else:
            for _, v in sub_schema.items():
                if isinstance(v, dict):
                    recursive_search(v, key)

    recursive_search(schema, key)
    return schema



"""Reads the content of multiples files from sim_telarray."""

import copy
import logging

import numpy as np
from ctapipe.io import write_table
from eventio import EventIOFile, Histograms
from eventio.search_utils import yield_toplevel_of_type

from simtools import version
from simtools.io_operations.hdf5_handler import fill_hdf5_table
from simtools.simtel.simtel_io_histogram import (
    HistogramIdNotFoundError,
    InconsistentHistogramFormatError,
    SimtelIOHistogram,
)
from simtools.utils.names import sanitize_name

__all__ = [
    "SimtelIOHistograms",
]


class SimtelIOHistograms:
    """
    Read the content of either multiple histogram (.hdata, or .hdata.zst) or simtel_array files.

    Allow both the .hdata.zst histogram and the .simtel.zst output file type.
    It uses the SimtelIOHistogram class to deal with individual files.
    Histogram files are ultimately handled by using eventio library.

    Parameters
    ----------
    histogram_files: list
        List of sim_telarray histogram files (str of Path).
    test: bool
        If True, only a fraction of the histograms will be processed, leading to a much shorter\
         runtime.
    area_from_distribution: bool
        If true, the area thrown (the area in which the simulated events are distributed)
        in the trigger rate calculation is estimated based on the event distribution.
        The expected shape of the distribution of events as function of the core distance is
        triangular up to the maximum distance. The weighted mean radius of the triangular
        distribution is 2/3 times the upper edge. Therefore, when using the
        ``area_from_distribution`` flag, the mean distance times 3/2, returns just the position of
        the upper edge in the triangle distribution with little impact of the binning and little
        dependence on the scatter area defined in the simulation. This is special useful when
        calculating trigger rate for individual telescopes.
        If false, the area thrown is estimated based on the maximum distance as given in
        the simulation configuration.
    energy_range: list
        The energy range used in the simulation. It must be passed as a list of floats and the
        energy must be in TeV (as in the CORSIKA configuration).
        This argument is only needed and used if histogram_file is a .hdata file, in which case the
        energy range cannot be retrieved directly from the file.
    view_cone: list
        The view cone used in the simulation. It must be passed as a list of floats and the
        view cone must be in deg (as in the CORSIKA configuration).
        This argument is only needed and used if histogram_file is a .hdata file, in which case the
        view cone cannot be retrieved directly from the file.
    """

    def __init__(
        self,
        histogram_files,
        test=False,
        area_from_distribution=False,
        energy_range=None,
        view_cone=None,
    ):
        """Initialize SimtelIOHistograms."""
        self._logger = logging.getLogger(__name__)
        if not isinstance(histogram_files, list):
            histogram_files = [histogram_files]
        self.histogram_files = histogram_files
        self.view_cone = view_cone
        self.energy_range = energy_range
        self._is_test = test
        self._combined_hists = None
        self._list_of_histograms = None
        self.__meta_dict = None
        self.area_from_distribution = area_from_distribution

    def calculate_trigger_rates(self, print_info=False, stack_files=False):
        """
        Calculate the triggered and simulated event rate considering the histograms in each file.

        It returns also a list with the tables where the energy dependent trigger rate for each
        file can be found.

        Parameters
        ----------
        print_info: bool
            if True, prints out the information about the histograms such as energy range, area,
            etc.
        stack_files: bool
            if True, stack the histograms from the different files into single histograms.
            Useful to increase event statistics when calculating the trigger rate.

        Returns
        -------
        sim_event_rates: list of astropy.Quantity[1/time]
            The simulated event rates.
        triggered_event_rates: list of astropy.Quantity[1/time]
            The triggered event rates.
        triggered_event_rate_uncertainties: list of astropy.Quantity[1/time]
            The uncertainties in the triggered event rates.
        trigger_rate_in_tables: list of astropy.QTable
            The energy dependent trigger rates.
            Only filled if stack_files is False.
        """
        if stack_files:
            (
                sim_event_rates,
                triggered_event_rates,
                triggered_event_rate_uncertainties,
            ) = self._rates_for_stacked_files()
            trigger_rate_in_tables = []
        else:
            (
                sim_event_rates,
                triggered_event_rates,
                triggered_event_rate_uncertainties,
                trigger_rate_in_tables,
            ) = self._rates_for_each_file(print_info)

        return (
            sim_event_rates,
            triggered_event_rates,
            triggered_event_rate_uncertainties,
            trigger_rate_in_tables,
        )

    def _fill_stacked_events(self):
        """
        Retrieve the simulated and triggered event histograms from the stacked histograms instead.

        Returns
        -------
        first_hist_file: dict
            The simulated 2D event histogram.
        second_hist_file: dict
            The triggered 2D event histogram.

        Raises
        ------
        HistogramIdNotFoundError:
            if histogram ids not found. Problem with the file.
        """
        sim_hist = None
        trig_hist = None
        for _, one_hist in enumerate(self.combined_hists):
            if one_hist["id"] == 1:
                sim_hist = one_hist
            elif one_hist["id"] == 2:
                trig_hist = one_hist

        if sim_hist is None or trig_hist is None:
            msg = (
                "Simulated and triggered histograms were not found in the stacked histograms."
                " Please check your simtel_array files!"
            )
            self._logger.error(msg)
            raise HistogramIdNotFoundError

        return sim_hist, trig_hist

    def get_stacked_num_events(self):
        """
        Return stacked number of simulated events and triggered events.

        Returns
        -------
        int:
            total number of simulated events for the stacked dataset.
        int:
            total number of triggered events for the stacked dataset.
        """
        stacked_num_simulated_events = 0
        stacked_num_triggered_events = 0
        for _, file in enumerate(self.histogram_files):
            simtel_hist_instance = SimtelIOHistogram(
                file,
                area_from_distribution=self.area_from_distribution,
                energy_range=self.energy_range,
                view_cone=self.view_cone,
            )
            _simulated, _triggered = simtel_hist_instance.total_number_of_events
            stacked_num_simulated_events += _simulated
            stacked_num_triggered_events += _triggered
        return stacked_num_simulated_events, stacked_num_triggered_events

    def _rates_for_stacked_files(self):
        """
        Calculate trigger rate for the stacked case.

        Returns
        -------
        sim_event_rates: list of astropy.Quantity[1/time]
            The simulated event rates.
        triggered_event_rates: list of astropy.Quantity[1/time]
            The triggered event rates.
        triggered_event_rate_uncertainties: list of astropy.Quantity[1/time]
            The uncertainties in the triggered event rates.
        trigger_rate_in_tables: list of astropy.QTable
            The energy dependent trigger rates.
            Only filled if stack_files is False.
        """
        logging.info("Estimates for the stacked histograms:")
        sim_hist, trig_hist = self._fill_stacked_events()
        # Using a dummy instance of SimtelIOHistogram to calculate the trigger rate for the
        # stacked files
        simtel_hist_instance = SimtelIOHistogram(
            self.histogram_files[0],
            area_from_distribution=self.area_from_distribution,
            energy_range=self.energy_range,
            view_cone=self.view_cone,
        )

        stacked_num_simulated_events, stacked_num_triggered_events = self.get_stacked_num_events()
        logging.info(f"Total number of simulated events: {stacked_num_simulated_events} events")
        logging.info(f"Total number of triggered events: {stacked_num_triggered_events} events")
        obs_time = simtel_hist_instance.estimate_observation_time(stacked_num_simulated_events)
        logging.info(
            "Estimated equivalent observation time corresponding to the number of"
            f"events simulated: {obs_time.value} s"
        )
        sim_event_rate = stacked_num_simulated_events / obs_time
        logging.info(f"Simulated event rate: {sim_event_rate.value:.4e} Hz")
        (
            triggered_event_rate,
            _,
        ) = simtel_hist_instance.compute_system_trigger_rate(
            events_histogram=sim_hist, triggered_events_histogram=trig_hist
        )
        triggered_event_rate_uncertainty = simtel_hist_instance.estimate_trigger_rate_uncertainty(
            triggered_event_rate, stacked_num_simulated_events, stacked_num_triggered_events
        )
        logging.info(
            f"System trigger event rate for stacked files: "
            f"{triggered_event_rate.value:.4e} \u00b1 "
            f"{triggered_event_rate_uncertainty.value:.4e} Hz"
        )
        return (
            [sim_event_rate],
            [triggered_event_rate],
            [triggered_event_rate_uncertainty],
        )

    def _rates_for_each_file(self, print_info=False):
        """
        Calculate trigger rate for each file.

        Returns
        -------
        sim_event_rates: list of astropy.Quantity[1/time]
            The simulated event rates.
        triggered_event_rates: list of astropy.Quantity[1/time]
            The triggered event rates.
        triggered_event_rate_uncertainties: list of astropy.Quantity[1/time]
            The uncertainties in the triggered event rates.
        """
        triggered_event_rates = []
        sim_event_rates = []
        trigger_rate_in_tables = []
        triggered_event_rate_uncertainties = []
        for i_file, file in enumerate(self.histogram_files):
            simtel_hist_instance = SimtelIOHistogram(
                file,
                area_from_distribution=self.area_from_distribution,
                energy_range=self.energy_range,
                view_cone=self.view_cone,
            )
            if print_info:
                simtel_hist_instance.print_info()

            _simulated_events, _triggered_events = simtel_hist_instance.total_number_of_events
            logging.info(f"Histogram {i_file + 1}:")
            logging.info(f"Total number of simulated events: {_simulated_events} events")
            logging.info(f"Total number of triggered events: {_triggered_events} events")

            obs_time = simtel_hist_instance.estimate_observation_time(_simulated_events)
            logging.info(
                f"Estimated equivalent observation time corresponding to the number of "
                f"events simulated: {obs_time.value} s"
            )
            sim_event_rate = _simulated_events / obs_time
            sim_event_rates.append(sim_event_rate)
            logging.info(f"Simulated event rate: {sim_event_rate.value:.4e} Hz")

            (
                triggered_event_rate,
                triggered_event_rate_uncertainty,
            ) = simtel_hist_instance.compute_system_trigger_rate()
            logging.info(
                f"System trigger event rate: "
                # pylint: disable=E1101
                f"{triggered_event_rate.value:.4e} \u00b1 "
                # pylint: disable=E1101
                f"{triggered_event_rate_uncertainty.value:.4e} Hz"
            )
            triggered_event_rates.append(triggered_event_rate)
            triggered_event_rate_uncertainties.append(triggered_event_rate_uncertainty)
            trigger_rate_in_tables.append(simtel_hist_instance.trigger_info_in_table())
        return (
            sim_event_rates,
            triggered_event_rates,
            triggered_event_rate_uncertainties,
            trigger_rate_in_tables,
        )

    @property
    def number_of_files(self):
        """Returns number of histograms."""
        return len(self.histogram_files)

    def _check_consistency(self, first_hist_file, second_hist_file):
        """
        Check whether two histograms have the same format.

        Raises an error in case they are not consistent.

        Parameters
        ----------
        first_hist_file: dict
            One histogram from a single file.
        second_hist_file: dict
            One histogram from a single file.

        Raises
        ------
        InconsistentHistogramFormatError:
            if the format of the histograms have inconsistent dimensions.
        """
        for key_to_test in [
            "lower_x",
            "upper_x",
            "n_bins_x",
            "title",
        ]:
            if first_hist_file[key_to_test] != second_hist_file[key_to_test]:
                msg = "Trying to add histograms with inconsistent dimensions"
                self._logger.error(msg)
                raise InconsistentHistogramFormatError

    @property
    def list_of_histograms(self):
        """
        Returns a list with the histograms for each file.

        Returns
        -------
        list:
            List of histograms.
        """
        if self._list_of_histograms is None:
            self._list_of_histograms = []
            for file in self.histogram_files:
                with EventIOFile(file) as f:
                    for o in yield_toplevel_of_type(f, Histograms):
                        hists = o.parse()
                        self._list_of_histograms.append(hists)
        return self._list_of_histograms

    @property
    def combined_hists(self):
        """
        Combine histograms of same type of histogram.

        Histograms are read from various lists into a single histogram list.
        """
        # Processing and combining histograms from multiple files
        if self._combined_hists is None:
            self._combined_hists = []
            for histogram_index, hists_one_file in enumerate(self.list_of_histograms):
                if histogram_index == 0:
                    # First file
                    self._combined_hists = copy.copy(hists_one_file)

                else:
                    for hist, this_combined_hist in zip(hists_one_file, self._combined_hists):
                        self._check_consistency(hist, this_combined_hist)

                        this_combined_hist["data"] = np.add(
                            this_combined_hist["data"], hist["data"]
                        )

            self._logger.debug(f"End of reading {len(self.histogram_files)} files")
        return self._combined_hists

    @combined_hists.setter
    def combined_hists(self, new_combined_hists):
        """
        Setter for combined_hists.

        Parameters
        ----------
        new_combined_hists:
            Combined histograms.
        """
        self._combined_hists = new_combined_hists

    def plot_one_histogram(self, histogram_index, ax):
        """
        Plot a single histogram referent to the index histogram_index.

        Parameters
        ----------
        histogram_index: int
            Index of the histogram to be plotted.
        ax: matplotlib.axes.Axes
            Instance of matplotlib.axes.Axes in which to plot the histogram.
        """
        hist = self.combined_hists[histogram_index]
        ax.set_title(hist["title"])

        def _get_bins(hist, axis=0):
            ax_str = "x" if axis == 0 else "y"
            return np.linspace(
                hist["lower_" + ax_str],
                hist["upper_" + ax_str],
                hist["n_bins_" + ax_str] + 1,
            )

        def _get_ax_lim(hist, axis=0):
            if np.sum(hist["data"]) == 0:
                return 0, 1

            bins = _get_bins(hist, axis=axis)

            if hist["data"].ndim == 1:
                non_zero = np.where(hist["data"] != 0)
            else:
                marginal = np.sum(hist["data"], axis=axis)
                non_zero = np.where(marginal != 0)

            return bins[non_zero[0][0]], bins[non_zero[0][-1] + 1]

        if hist["n_bins_y"] > 0:
            # 2D histogram

            xlim = _get_ax_lim(hist, axis=0)
            ylim = _get_ax_lim(hist, axis=1)

            if np.sum(hist["data"]) == 0:
                ax.text(
                    0.5,
                    0.5,
                    "EMPTY",
                    horizontalalignment="center",
                    verticalalignment="center",
                    transform=ax.transAxes,
                )
                return

            x_bins = _get_bins(hist, axis=0)
            y_bins = _get_bins(hist, axis=1)

            ax.pcolormesh(x_bins, y_bins, hist["data"])
            ax.set_xlim(xlim)
            ax.set_ylim(ylim)

        else:
            # 1D histogram

            xlim = _get_ax_lim(hist, axis=0)

            if np.sum(hist["data"]) == 0:
                ax.text(
                    0.5,
                    0.5,
                    "EMPTY",
                    horizontalalignment="center",
                    verticalalignment="center",
                    transform=ax.transAxes,
                )
                return

            x_bins = _get_bins(hist, axis=0)
            centers = 0.5 * (x_bins[:-1] + x_bins[1:])
            ax.hist(centers, bins=x_bins, weights=hist["data"])
            ax.set_xlim(xlim)

    @property
    def _meta_dict(self):
        """
        Define the meta dictionary for exporting the histograms.

        Returns
        -------
        dict
            Meta dictionary for the hdf5 files with the histograms.
        """
        if self.__meta_dict is None:
            self.__meta_dict = {
                "simtools_version": version.__version__,
                "note": "Only lower bin edges are given.",
            }
        return self.__meta_dict

    def export_histograms(self, hdf5_file_name, overwrite=False):
        """
        Export sim_telarray histograms to hdf5 files.

        Parameters
        ----------
        hdf5_file_name: str
            Name of the file to be saved with the hdf5 tables.
        overwrite: bool
            If True overwrites histograms already saved in the hdf5 file.
        """
        self._logger.info(f"Exporting histograms to {hdf5_file_name}.")
        for histogram in self.combined_hists:
            x_bin_edges_list = np.linspace(
                histogram["lower_x"],
                histogram["upper_x"],
                num=histogram["n_bins_x"] + 1,
                endpoint=True,
            )
            if histogram["n_bins_y"] > 0:
                y_bin_edges_list = np.linspace(
                    histogram["lower_y"],
                    histogram["upper_y"],
                    num=histogram["n_bins_y"] + 1,
                    endpoint=True,
                )
            else:
                y_bin_edges_list = None

            self._meta_dict["Title"] = sanitize_name(histogram["title"])

            table = fill_hdf5_table(
                hist=histogram["data"],
                x_bin_edges=x_bin_edges_list,
                y_bin_edges=y_bin_edges_list,
                x_label=None,
                y_label=None,
                meta_data=self._meta_dict,
            )

            self._logger.debug(
                f"Writing histogram with name {self._meta_dict['Title']} to {hdf5_file_name}."
            )
            # overwrite takes precedence over append
            write_table(
                table,
                hdf5_file_name,
                f"/{self._meta_dict['Title']}",
                append=not overwrite,
                overwrite=overwrite,
            )



"""Definition and modeling of mirror panels."""

import logging
from pathlib import Path

import astropy.io.ascii
import astropy.units as u
import numpy as np
from astropy.table import Table

__all__ = ["InvalidMirrorListFileError", "Mirrors"]


class InvalidMirrorListFileError(Exception):
    """Exception for invalid mirror list file."""


class Mirrors:
    """
    Mirrors class, created from a mirror list file.

    Parameters
    ----------
    mirror_list_file: Union[str, Path]
        Mirror list in sim_telarray or ecsv format (with panel focal length only).
    parameters: dict, optional
        Dictionary of parameters from the database.
    """

    def __init__(self, mirror_list_file: str | Path, parameters: dict | None = None):
        """Initialize Mirrors."""
        self._logger = logging.getLogger(__name__)
        self._logger.debug("Mirrors Init")

        self.mirror_table = Table()
        self.mirror_diameter = None
        self.shape_type = None
        self.number_of_mirrors = 0
        self.parameters = parameters

        self._mirror_list_file = mirror_list_file
        self._read_mirror_list()

    def _read_mirror_list(self):
        """
        Read the mirror lists from disk and store the data.

        Allow reading of mirror lists in sim_telarray and ecsv format.
        """
        if str(self._mirror_list_file).find("ecsv") > 0:
            self._read_mirror_list_from_ecsv()
        else:
            self._read_mirror_list_from_sim_telarray()

    def _read_mirror_list_from_ecsv(self):
        """
        Read the mirror list in ecsv format and store the data.

        Raises
        ------
        InvalidMirrorListFileError
            If number of mirrors is 0.
        """
        self._logger.debug(f"Reading mirror properties from {self._mirror_list_file}")
        self.mirror_table = Table.read(self._mirror_list_file, format="ascii.ecsv")

        self.number_of_mirrors = np.shape(self.mirror_table)[0]
        self._logger.debug(f"Number of Mirrors = {self.number_of_mirrors}")

        if self.number_of_mirrors == 0:
            msg = "Problem reading mirror list file"
            self._logger.error(msg)
            raise InvalidMirrorListFileError

        try:
            self.mirror_diameter = u.Quantity(self.mirror_table["mirror_diameter"])[0]
            self._logger.debug(f"Mirror diameter = {self.mirror_diameter}")
        except KeyError:
            self._logger.debug("Mirror mirror_panel_diameter not in mirror file")
            try:
                self.mirror_diameter = u.Quantity(
                    self.parameters["mirror_panel_diameter"]["value"],
                    self.parameters["mirror_panel_diameter"]["unit"],
                )
                self._logger.debug("Take mirror_panel_diameter from parameters")
            except TypeError as error:
                msg = "Mirror mirror_panel_diameter not contained in DB"
                self._logger.error(msg)
                raise TypeError(msg) from error
        if "focal_length" not in self.mirror_table.colnames:
            try:
                self.mirror_table["focal_length"] = (
                    self.mirror_table["mirror_curvature_radius"].to("cm") / 2
                )
            except KeyError:
                self._logger.debug("mirror_curvature_radius not contained in mirror list")
                try:
                    self.mirror_table["focal_length"] = self.number_of_mirrors * [
                        u.Quantity(
                            self.parameters["mirror_focal_length"]["value"],
                            self.parameters["mirror_focal_length"]["unit"],
                        )
                    ]
                    self._logger.debug("Take mirror_focal_length from parameters")
                except TypeError as error:
                    msg = "mirror_focal_length not contained in DB"
                    self._logger.error(msg)
                    raise TypeError(msg) from error

        try:
            self.shape_type = u.Quantity(self.mirror_table["shape_type"])[0]
            self._logger.debug(f"Mirror shape_type = {self.shape_type}")
        except KeyError:
            self._logger.debug("Mirror shape_type not in mirror file")
            try:
                self.shape_type = self.parameters["mirror_panel_shape"]["value"]
                self._logger.debug("Take shape_type from parameters")
            except TypeError as error:
                msg = "Mirror shape_type not contained in DB"
                self._logger.error(msg)
                raise TypeError(msg) from error

    def _read_mirror_list_from_sim_telarray(self):
        """
        Read the mirror list in sim_telarray format and store the data.

        Allow to read mirror lists with different number of columns.

        Raises
        ------
        InvalidMirrorListFileError
            If number of mirrors is 0.
        """
        self._logger.debug(f"Reading mirror properties from {self._mirror_list_file}")

        try:
            self.mirror_table = Table.read(
                self._mirror_list_file,
                format="ascii.no_header",
                names=[
                    "mirror_x",
                    "mirror_y",
                    "mirror_diameter",
                    "focal_length",
                    "shape_type",
                    "mirror_z",
                    "sep",
                    "mirror_panel_id",
                ],
                units=["cm", "cm", "cm", "cm", None, "cm", None, None],
            )
            self.mirror_table["mirror_panel_id"] = np.array(
                [
                    int("".join(filter(str.isdigit, string)))
                    for string in self.mirror_table["mirror_panel_id"]
                ]
            )
        except astropy.io.ascii.core.InconsistentTableError:
            self._logger.debug("Try and read mirror list with low number of columns")
            self.mirror_table = Table.read(
                self._mirror_list_file,
                format="ascii.no_header",
                names=[
                    "mirror_x",
                    "mirror_y",
                    "mirror_diameter",
                    "focal_length",
                    "shape_type",
                ],
                units=["cm", "cm", "cm", "cm", None],
            )
            self.mirror_table["mirror_panel_id"] = np.arange(len(self.mirror_table["mirror_x"]))

        self.shape_type = self.mirror_table["shape_type"][0]
        self.mirror_diameter = u.Quantity(
            self.mirror_table["mirror_diameter"][0], self.mirror_table["mirror_diameter"].unit
        )
        self.number_of_mirrors = len(self.mirror_table["focal_length"])

        self._logger.debug(f"Mirror shape_type = {self.shape_type}")
        self._logger.debug(f"Mirror diameter = {self.mirror_diameter}")
        self._logger.debug(f"Number of Mirrors = {self.number_of_mirrors}")

    def get_single_mirror_parameters(self, number: int) -> tuple:
        """
        Get parameters for a single mirror given by number.

        Parameters
        ----------
        number: int
            Mirror number of desired parameters.

        Returns
        -------
        tuple
            (pos_x, pos_y, mirror_diameter, focal_length, shape_type): tuple of float
            X, Y positions, mirror_diameter, focal length and shape_type.
        """
        mask = self.mirror_table["mirror_panel_id"] == number
        if not np.any(mask):
            self._logger.debug(f"Mirror id{number} not in table, using first mirror instead")
            mask[0] = True
        try:
            return_values = (
                u.Quantity(
                    self.mirror_table[mask]["mirror_x"].value[0],
                    self.mirror_table[mask]["mirror_x"].unit,
                ),
                u.Quantity(
                    self.mirror_table[mask]["mirror_y"].value[0],
                    self.mirror_table[mask]["mirror_y"].unit,
                ),
                u.Quantity(
                    self.mirror_table[mask]["mirror_diameter"].value[0],
                    self.mirror_table[mask]["mirror_diameter"].unit,
                ),
                u.Quantity(
                    self.mirror_table[mask]["focal_length"].value[0],
                    self.mirror_table[mask]["focal_length"].unit,
                ),
                u.Quantity(
                    self.mirror_table[mask]["shape_type"].value[0],
                    self.mirror_table[mask]["shape_type"].unit,
                ),
            )
        except KeyError:
            self._logger.debug("Mirror list missing required column")
            return_values = (
                0,
                0,
                self.mirror_diameter,
                u.Quantity(
                    self.mirror_table[mask]["focal_length"].value[0],
                    self.mirror_table[mask]["focal_length"].unit,
                ),
                self.shape_type,
            )
        return return_values

    def plot_mirror_layout(self):
        """Plot the mirror layout (not implemented yet)."""



#!/usr/bin/python3
"""Helper functions calculations related to model parameters."""

import math

from simtools.utils import names

__all__ = [
    "compute_telescope_transmission",
    "is_two_mirror_telescope",
]


def compute_telescope_transmission(pars: list[float], off_axis: float) -> float:
    """
    Compute telescope transmission (0 < T < 1) for a given off-axis angle.

    The telescope transmission depends on the MC model used.

    Parameters
    ----------
    pars: list of float
        Parameters of the telescope transmission. Len(pars) should be 5 or 6.
    off_axis: float
        Off-axis angle in deg.

    Returns
    -------
    float
        Telescope transmission.
    """
    _deg_to_rad = math.pi / 180.0
    if pars[1] == 0:
        return pars[0]

    t = math.sin(off_axis * _deg_to_rad) / (pars[3] * _deg_to_rad)
    return pars[0] / (1.0 + pars[2] * t ** pars[4])


def is_two_mirror_telescope(telescope_model_name: str) -> bool:
    """
    Determine if the telescope model is a two-mirror telescope.

    Parameters
    ----------
    telescope_model_name: str
        Name of the telescope model.

    Returns
    -------
    bool
        True if it is a two-mirror telescope.
    """
    tel_type = names.get_array_element_type_from_name(telescope_model_name)
    if "SST" in tel_type or "SCT" in tel_type:
        return True
    return False



"""Base class for running sim_telarray simulations."""

import logging
import stat
import subprocess
from pathlib import Path

import simtools.utils.general as gen
from simtools.runners.runner_services import RunnerServices

__all__ = ["InvalidOutputFileError", "SimtelExecutionError", "SimtelRunner"]


class SimtelExecutionError(Exception):
    """Exception for simtel_array execution error."""


class InvalidOutputFileError(Exception):
    """Exception for invalid output file."""


class SimtelRunner:
    """
    Base class for running sim_telarray simulations.

    Parameters
    ----------
    simtel_path: str or Path
        Location of sim_telarray installation.
    label: str
        Instance label. Important for output file naming.
    """

    def __init__(self, simtel_path, label=None, corsika_config=None, use_multipipe=False):
        """Initialize SimtelRunner."""
        self._logger = logging.getLogger(__name__)

        self._simtel_path = Path(simtel_path)
        self.label = label
        self._base_directory = None

        self.runs_per_set = 1

        self.runner_service = RunnerServices(corsika_config, label)
        self._directory = self.runner_service.load_data_directories(
            "corsika_simtel" if use_multipipe else "simtel"
        )

    def __repr__(self):
        """Return a string representation of the SimtelRunner object."""
        return f"SimtelRunner(label={self.label})\n"

    def prepare_run_script(self, test=False, input_file=None, run_number=None, extra_commands=None):
        """
        Build and return the full path of the bash run script containing the sim_telarray command.

        Parameters
        ----------
        test: bool
            Test flag for faster execution.
        input_file: str or Path
            Full path of the input CORSIKA file.
        run_number: int
            Run number.
        extra_commands: str
            Additional commands for running simulations given in config.yml.

        Returns
        -------
        Path
            Full path of the run script.
        """
        self._logger.debug("Creating run bash script")

        script_file_path = self.get_file_name(file_type="sub_script", run_number=run_number)

        self._logger.debug(f"Run bash script - {script_file_path}")

        self._logger.debug(f"Extra commands to be added to the run script {extra_commands}")

        command = self._make_run_command(run_number=run_number, input_file=input_file)
        with script_file_path.open("w", encoding="utf-8") as file:
            file.write("#!/usr/bin/env bash\n\n")
            file.write("set -e\n")
            file.write("set -o pipefail\n")
            file.write("\nSECONDS=0\n")

            if extra_commands is not None:
                file.write("# Writing extras\n")
                for line in extra_commands:
                    file.write(f"{line}\n")
                file.write("# End of extras\n\n")

            n = 1 if test else self.runs_per_set
            for _ in range(n):
                file.write(f"{command}\n\n")

            file.write('\necho "RUNTIME: $SECONDS"\n')

        script_file_path.chmod(script_file_path.stat().st_mode | stat.S_IXUSR | stat.S_IXGRP)
        return script_file_path

    def run(self, test=False, input_file=None, run_number=None):
        """
        Make run command and run sim_telarray.

        Parameters
        ----------
        test: bool
            If True, make simulations faster.
        input_file: str or Path
            Full path of the input CORSIKA file.
        run_number: int
            Run number.
        """
        self._logger.debug("Running sim_telarray")

        command, stdout_file, stderr_file = self._make_run_command(
            run_number=run_number, input_file=input_file
        )

        if test:
            self._logger.info(f"Running (test) with command: {command}")
            self._run_simtel_and_check_output(command, stdout_file, stderr_file)
        else:
            self._logger.debug(f"Running ({self.runs_per_set}x) with command: {command}")
            for _ in range(self.runs_per_set):
                self._run_simtel_and_check_output(command, stdout_file, stderr_file)

        self._check_run_result(run_number=run_number)

    def _check_run_result(self, run_number=None):  # pylint: disable=all
        """Check if simtel output file exists."""
        pass

    def _raise_simtel_error(self):
        """
        Raise sim_telarray execution error.

        Final 30 lines from the log file are collected and printed.

        Raises
        ------
        SimtelExecutionError
        """
        if hasattr(self, "_log_file"):
            msg = gen.get_log_excerpt(self._log_file)
        else:
            msg = "Simtel log file does not exist."

        self._logger.error(msg)
        raise SimtelExecutionError(msg)

    def _run_simtel_and_check_output(self, command, stdout_file, stderr_file):
        """
        Run the sim_telarray command and check the exit code.

        Raises
        ------
        SimtelExecutionError
            if run was not successful.
        """
        stdout_file = stdout_file if stdout_file else "/dev/null"
        stderr_file = stderr_file if stderr_file else "/dev/null"
        with (
            open(f"{stdout_file}", "w", encoding="utf-8") as stdout,
            open(f"{stderr_file}", "w", encoding="utf-8") as stderr,
        ):
            result = subprocess.run(
                command,
                shell=True,
                text=True,
                stdout=stdout,
                stderr=stderr,
            )

        if result.returncode != 0:
            self._logger.error(result.stderr)
            self._raise_simtel_error()
        return result.returncode

    def _make_run_command(self, run_number=None, input_file=None):
        self._logger.debug(
            "make_run_command is being called from the base class - "
            "it should be implemented in the sub class"
        )
        input_file = input_file if input_file else "nofile"
        run_number = run_number if run_number else 1
        return f"{input_file}-{run_number}", None, None

    @staticmethod
    def get_config_option(par, value=None, weak_option=False):
        """
        Build sim_telarray command.

        Parameters
        ----------
        par: str
            Parameter name.
        value: str
            Parameter value.
        weak_option: bool
            If True, use -W option instead of -C.

        Returns
        -------
        str
            Command for sim_telarray.
        """
        option_syntax = "-W" if weak_option else "-C"
        c = f" {option_syntax} {par}"
        c += f"={value}" if value is not None else ""
        return c

    def get_resources(self, run_number=None):
        """Return computing resources used."""
        return self.runner_service.get_resources(run_number)

    def get_file_name(self, simulation_software="simtel", file_type=None, run_number=None, mode=""):
        """
        Get the full path of a file for a given run number.

        Parameters
        ----------
        simulation_software: str
            Simulation software.
        file_type: str
            File type.
        run_number: int
            Run number.

        Returns
        -------
        str
            File name with full path.
        """
        if simulation_software.lower() != "simtel":
            raise ValueError(
                f"simulation_software ({simulation_software}) is not supported in SimulatorArray"
            )
        return self.runner_service.get_file_name(
            file_type=file_type, run_number=run_number, mode=mode
        )



#!/usr/bin/python3

r"""
Derive simulation configuration parameters for a simulation production.

Derived simulation configuration parameters include:

* energy range
* shower core scatter radius
* view cone radius
* total number of events to be simulated

Configuration parameters depend on characteristics of the observations, especially elevation,
azimuth, and night sky background.

The configuration parameters are derived according to the required precision. The metrics are:

* statistical uncertainty on the determination of the effective area as function of primary energy
* fraction of lost events to the selected core scatter and view cone radius (to be implemented)
* statistical uncertainty of the energy migration matrix as function of primary energy
    (to be implemented)

Command line arguments
----------------------
azimuth (float, required)
        Azimuth angle in degrees.
elevation (float, required)
        Elevation angle in degrees.
nsb (float, required)
        Night sky background value.
file_path (str, required)
        Path to file with MC events at CTAO DL2 data level.
        Used for statistical uncertainty evaluation.
file_type (str, required)
        Type of the DL2 MC event file ('point-like' or 'cone').
metrics (str, required)
        Path to a YAML file containing metrics for evaluation.
site (str, required)
        The observatory site (North or South).

Example
-------
To run the simulation configuration, execute the script as follows:

.. code-block:: console

        simtools-production-generate-simulation-config --azimuth 60.0 --elevation 45.0 \
                --nsb 0.3 --file_path tests/resources/production_dl2_fits/dl2_mc_events_file.fits \
                --file_type "point-like"    \
                --metrics_file tests/resources/production_simulation_config_metrics.yml --site North

The output will show the derived simulation parameters.
"""

import json
import logging
from pathlib import Path

import astropy.units as u

import simtools.utils.general as gen
from simtools.configuration import configurator
from simtools.data_model import schema
from simtools.io_operations import io_handler
from simtools.production_configuration.generate_simulation_config import (
    SimulationConfig,
)


def _parse(label):
    """Parse command-line arguments."""
    config = configurator.Configurator(
        label=label,
        description="Derive simulation configuration parameters for a simulation production.",
    )
    config.parser.add_argument(
        "--azimuth", type=float, required=True, help="Azimuth angle in degrees."
    )
    config.parser.add_argument(
        "--elevation", type=float, required=True, help="Elevation angle in degrees."
    )
    config.parser.add_argument(
        "--nsb", type=float, required=True, help="Night sky background in units of 1/(sr*ns*cm**2)."
    )
    config.parser.add_argument(
        "--file_path", type=str, required=True, help="Path to MC event file in DL2 format."
    )
    config.parser.add_argument(
        "--file_type",
        type=str,
        required=True,
        help="Type of the DL2 MC event file ('point-like' or 'cone').",
    )
    config.parser.add_argument(
        "--metrics_file",
        required=True,
        type=str,
        help="Path to YAML file containing metrics and required precision as values.",
    )
    config.parser.add_argument(
        "--site",
        type=str,
        required=True,
        help="Site location for the simulation (e.g., 'North', 'South').",
    )
    config.parser.add_argument(
        "--output_file",
        help="Name of the file to save the configured simulation parameters.",
        type=str,
        required=False,
        default="configured_simulation_params.json",
    )

    return config.initialize(db_config=False)


def main():
    """Run the Simulation Config application."""
    label = Path(__file__).stem
    args_dict, _ = _parse(label)

    logger = logging.getLogger()
    logger.setLevel(gen.get_log_level_from_user(args_dict["log_level"]))
    output_path = io_handler.IOHandler().get_output_directory(label)
    output_filepath = Path(output_path).joinpath(f"{args_dict['output_file']}")

    grid_point_config = {
        "azimuth": args_dict["azimuth"],
        "elevation": args_dict["elevation"],
        "night_sky_background": args_dict["nsb"],
    }

    metrics = gen.collect_data_from_file(args_dict["metrics_file"])
    schema.validate_dict_using_schema(
        data=metrics, schema_file="production_configuration_metrics.schema.yml"
    )

    simulation_config = SimulationConfig(
        grid_point=grid_point_config,
        file_path=args_dict["file_path"],
        file_type=args_dict["file_type"],
        metrics=metrics,
    )

    simulation_params = simulation_config.configure_simulation()

    serializable_config = {}

    for key, value in simulation_params.items():
        if isinstance(value, u.Quantity):
            serializable_config[key] = f"{value.value} {value.unit}"
        else:
            serializable_config[key] = value

    logger.info(f"Simulation configuration: {serializable_config}")
    with open(output_filepath, "w", encoding="utf-8") as f:
        json.dump(serializable_config, f, indent=4)
    logger.info(f"Simulation configuration saved to: {output_filepath}")


if __name__ == "__main__":
    main()



#!/usr/bin/python3

r"""
Application to run the StatisticalErrorEvaluator and interpolate results.

This application evaluates statistical uncertainties from DL2 MC event files
based on input parameters like zenith angles and offsets, and can perform interpolation
for a specified grid point.

Command line arguments
----------------------
base_path (str, required)
    Path to the directory containing the DL2 MC event file for interpolation.
zeniths (list of int, required)
    List of zenith angles to consider.
offsets (list of int, required)
    List of offsets in degrees.
interpolate (bool, optional)
    If set, performs interpolation for a specific grid point.
query_point (list of int, optional)
    Grid point for interpolation (energy, azimuth, zenith, NSB, offset).

Example
-------
To evaluate statistical uncertainties and perform interpolation, run the command line script:

.. code-block:: console

    simtools-production-scale-events --base_path tests/resources/production_dl2_fits/ \
        --zeniths 20 52 40 60 --offsets 0 --interpolate --query_point 1 180 30 0 0 \
        --metrics_file "path/to/metrics.yaml"


The output will display the scaled events for the specified grid point.
"""

import json
import logging
from pathlib import Path

import astropy.units as u
import numpy as np

import simtools.utils.general as gen
from simtools.configuration import configurator
from simtools.configuration.commandline_parser import CommandLineParser
from simtools.io_operations import io_handler
from simtools.production_configuration.calculate_statistical_errors_grid_point import (
    StatisticalErrorEvaluator,
)
from simtools.production_configuration.interpolation_handler import InterpolationHandler


def _parse(label, description):
    """
    Parse command line arguments for the statistical error evaluator application.

    Returns
    -------
    argparse.Namespace
        Parsed command line arguments.
    """
    config = configurator.Configurator(label=label, description=description)

    config.parser.add_argument(
        "--base_path",
        type=str,
        required=True,
        help="Path to the DL2 MC event files for interpolation.",
    )
    config.parser.add_argument(
        "--zeniths", nargs="+", type=CommandLineParser.zenith_angle, help="List of zenith angles."
    )
    config.parser.add_argument(
        "--offsets", nargs="+", type=float, help="List of offsets in degrees."
    )

    config.parser.add_argument(
        "--interpolate", action="store_true", help="Interpolate results for a specific grid point."
    )
    config.parser.add_argument(
        "--query_point",
        nargs=5,
        type=float,
        help="Grid point for interpolation (energy, azimuth, zenith, NSB, offset).",
    )
    config.parser.add_argument(
        "--output_file",
        type=str,
        default="interpolated_scaled_events.json",
        help="Output file to store the results. (default: 'interpolated_scaled_events.json').",
    )
    config.parser.add_argument(
        "--metrics_file",
        type=str,
        default="production_simulation_config_metrics.yml",
        help="Metrics definition file. (default: production_simulation_config_metrics.yml)",
    )
    return config.initialize(db_config=False)


def main():
    """Run the evaluator and interpolate."""
    label = Path(__file__).stem
    args_dict, _ = _parse(
        label,
        "Evaluate statistical uncertainties from DL2 MC event files and interpolate results.",
    )
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    logger.info(f"args dict: {args_dict}")

    output_path = io_handler.IOHandler().get_output_directory(label)
    output_filepath = Path(output_path).joinpath(f"{args_dict['output_file']}")

    evaluator_instances = []

    metrics = (
        gen.collect_data_from_file(args_dict["metrics_file"]) if "metrics_file" in args_dict else {}
    )

    if args_dict["base_path"] and args_dict["zeniths"] and args_dict["offsets"]:
        for zenith in args_dict["zeniths"]:
            for offset in args_dict["offsets"]:
                # Build file path based on base_path, zenith, and offset
                file_name = f"prod6_LaPalma-{int(zenith.value)}deg_"
                file_name += "gamma_cone.N.Am-4LSTs09MSTs_ID0_reduced.fits"
                file_path = Path(args_dict["base_path"]).joinpath(file_name)

                evaluator = StatisticalErrorEvaluator(
                    file_path,
                    file_type="Gamma-cone",
                    metrics=metrics,
                    grid_point=(1 * u.TeV, 180 * u.deg, zenith, 0, offset * u.deg),
                )

                evaluator.calculate_metrics()
                evaluator_instances.append(evaluator)

    else:
        logger.warning("No files read")
        logger.warning(f"Base Path: {args_dict['base_path']}")
        logger.warning(f"Zeniths: {args_dict['zeniths']}")
        logger.warning(f"Offsets: {args_dict['offsets']}")

    # Perform interpolation for the given query point
    interpolation_handler = InterpolationHandler(evaluator_instances, metrics=metrics)
    query_points = np.array([args_dict["query_point"]])
    scaled_events = interpolation_handler.interpolate(query_points)

    output_data = {
        "query_point": args_dict["query_point"],
        "scaled_events": scaled_events.tolist(),
    }
    with open(output_filepath, "w", encoding="utf-8") as f:
        json.dump(output_data, f, indent=4)
    logger.info(f"Output saved to {output_filepath}")
    logger.info(f"Scaled events for grid point {args_dict['query_point']}: {scaled_events}")


if __name__ == "__main__":
    main()



#!/usr/bin/python3

r"""
    Validate the optical model parameters through ray tracing simulations of the whole telescope.

    A point-like light source is assumed. The output includes PSF (D80), \
    effective mirror area and effective focal length as a function of the off-axis angle. \

    The telescope zenith angle and the source distance can be set by command line arguments.

    Examples of the plots generated by this application are shown below. On the top, the D80 \
    vs off-axis is shown in cm (left) and deg (right). On the bottom, the effective mirror \
    area (left) and the effective focal length (right) vs off-axis angle are shown.

    .. _validate_optics_plot:
    .. image::  images/validate_optics_North-LST-1_d80_cm.png
      :width: 49 %
    .. image::  images/validate_optics_North-LST-1_d80_deg.png
      :width: 49 %

    .. image::  images/validate_optics_North-LST-1_eff_area.png
      :width: 49 %
    .. image::  images/validate_optics_North-LST-1_eff_flen.png
      :width: 49 %


    Command line arguments
    ----------------------
    site (str, required)
        North or South.
    telescope (str, required)
        Telescope model name (e.g. LST-1, SST-D, ...).
    model_version (str, optional)
        Model version.
    src_distance (float, optional)
        Source distance in km.
    zenith (float, optional)
        Zenith angle in deg.
    max_offset (float, optional)
        Maximum offset angle in deg.
    offset_steps (float, optional)
        Offset angle step size.
    plot_images (activation mode, optional)
        Produce a multiple pages pdf file with the image plots.
    test (activation mode, optional)
        If activated, application will be faster by simulating fewer photons.

    Example
    -------
    LST-1 5.0.0

    .. code-block:: console

        simtools-validate-optics --site North --telescope LST-1 --max_offset 1.0 \\
        --zenith 20 --src_distance 10 --test

    The output is saved in simtools-output/validate_optics

    Expected final print-out message:

    .. code-block:: console

        INFO::ray_tracing(l434)::plot::Plotting eff_area vs off-axis angle
        INFO::ray_tracing(l434)::plot::Plotting eff_flen vs off-axis angle

"""

import logging
from pathlib import Path

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.backends.backend_pdf import PdfPages

import simtools.utils.general as gen
from simtools.configuration import configurator
from simtools.io_operations import io_handler
from simtools.model.telescope_model import TelescopeModel
from simtools.ray_tracing.ray_tracing import RayTracing
from simtools.visualization import visualize


def _parse(label):
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=label,
        description=(
            "Calculate and plot the PSF and effective mirror area as a function of off-axis angle "
            "of the telescope requested."
        ),
    )

    config.parser.add_argument(
        "--src_distance",
        help="Source distance in km",
        type=float,
        default=10,
    )
    config.parser.add_argument("--zenith", help="Zenith angle in deg", type=float, default=20)
    config.parser.add_argument(
        "--max_offset",
        help="Maximum offset angle in deg",
        type=float,
        default=4,
    )
    config.parser.add_argument(
        "--offset_steps",
        help="Offset angle step size",
        type=float,
        default=0.25,
    )
    config.parser.add_argument(
        "--plot_images",
        help="Produce a multiple pages pdf file with the image plots.",
        action="store_true",
    )
    return config.initialize(db_config=True, simulation_model=["telescope", "model_version"])


def main():  # noqa: D103
    label = Path(__file__).stem
    args_dict, db_config = _parse(label)

    logger = logging.getLogger()
    logger.setLevel(gen.get_log_level_from_user(args_dict["log_level"]))

    _io_handler = io_handler.IOHandler()
    output_dir = _io_handler.get_output_directory(label, sub_dir="application-plots")

    tel_model = TelescopeModel(
        site=args_dict["site"],
        telescope_name=args_dict["telescope"],
        model_version=args_dict["model_version"],
        label=label,
        mongo_db_config=db_config,
    )

    ######################################################################
    # This is here as an example how to change parameters when necessary.
    ######################################################################
    # pars_to_change = {
    #     'mirror_focal_length': 1608.3,
    #     'mirror_offset': -177.5,
    #     'camera_body_diameter': 289.7,
    #     'telescope_transmission': 1
    # }
    # tel_model.change_multiple_parameters(**pars_to_change)

    print(f"\nValidating telescope optics with ray tracing simulations for {tel_model.name}\n")

    ray = RayTracing(
        telescope_model=tel_model,
        simtel_path=args_dict["simtel_path"],
        zenith_angle=args_dict["zenith"] * u.deg,
        source_distance=args_dict["src_distance"] * u.km,
        off_axis_angle=np.linspace(
            0,
            args_dict["max_offset"],
            int(args_dict["max_offset"] / args_dict["offset_steps"]) + 1,
        )
        * u.deg,
    )
    ray.simulate(test=args_dict["test"], force=False)
    ray.analyze(force=True)

    # Plotting
    for key in ["d80_deg", "d80_cm", "eff_area", "eff_flen"]:
        plt.figure(figsize=(8, 6), tight_layout=True)

        ray.plot(key, marker="o", linestyle=":", color="k")

        plot_file_name = "_".join((label, tel_model.name, key))
        plot_file = output_dir.joinpath(plot_file_name)
        visualize.save_figure(plt, plot_file)

    # Plotting images
    if args_dict["plot_images"]:
        plot_file_name = "_".join((label, tel_model.name, "images.pdf"))
        plot_file = output_dir.joinpath(plot_file_name)
        pdf_pages = PdfPages(plot_file)

        logger.info(f"Plotting images into {plot_file}")

        for image in ray.images():
            fig = plt.figure(figsize=(8, 6), tight_layout=True)
            image.plot_image()
            pdf_pages.savefig(fig)
            plt.clf()
        plt.close()
        pdf_pages.close()


if __name__ == "__main__":
    main()



#!/usr/bin/python3

"""
Run simtools applications from configuration files for setting workflows.

Allows to run several simtools applications with a single configuration file, which includes
both the name of the simtools application and the configuration for the application.

Strong assumption on the directory structure for input and output files of applications.

"""

import logging
import subprocess
import tempfile
from pathlib import Path

import yaml

import simtools.utils.general as gen
from simtools import dependencies
from simtools.configuration import configurator


def _parse(label, description, usage):
    """
    Parse command line configuration.

    Parameters
    ----------
    label : str
        Label describing the application.
    description : str
        Description of the application.
    usage : str
        Example on how to use the application.

    Returns
    -------
    CommandLineParser
        Command line parser object.
    """
    config = configurator.Configurator(label=label, description=description, usage=usage)

    config.parser.add_argument(
        "--configuration_file",
        help="Application configuration.",
        type=str,
        required=True,
        default=None,
    )
    return config.initialize(db_config=True)


def run_application(application, configuration, logger):
    """Run a simtools application and return stdout and stderr."""
    with tempfile.NamedTemporaryFile(mode="w", delete=True, suffix=".yml") as temp_config:
        yaml.dump(configuration, temp_config, default_flow_style=False)
        temp_config.flush()
        configuration_file = Path(temp_config.name)
        try:
            result = subprocess.run(
                [application, "--config", configuration_file],
                check=True,
                capture_output=True,
                text=True,
            )
        except subprocess.CalledProcessError as exc:
            logger.error(f"Error running application {application}: {exc.stderr}")
            raise exc
        return result.stdout, result.stderr


def get_subdirectory_name(path):
    """Get the first subdirectory name under 'input'."""
    path = Path(path).resolve()
    try:
        input_index = path.parts.index("input")
        return path.parts[input_index], path.parts[input_index + 1]
    except (ValueError, IndexError) as exc:
        raise ValueError(f"Could not find subdirectory under 'input': {exc}") from exc


def read_application_configuration(configuration_file, logger):
    """
    Read application configuration from file and modify for setting workflows.

    Strong assumptions on the structure of input and output files:

    - configuration file is expected to be in './input/<workflow directory>/<yaml file>'
    - output files will be written out to './output/<workflow directory>/'

    Replaces the placeholders in the configuration file with the actual values.
    Sets 'USE_PLAIN_OUTPUT_PATH' to True for all applications.

    Parameters
    ----------
    configuration_file : str
        Configuration file name.
    logger : Logger
        Logger object.

    Returns
    -------
    dict
        Application configuration.
    Path
        Path to the log file.

    """
    application_config = gen.collect_data_from_file(configuration_file).get("CTA_SIMPIPE")
    place_holder = "__SETTING_WORKFLOW__"
    workflow_dir, setting_workflow = get_subdirectory_name(configuration_file)
    output_path = (
        workflow_dir.with_name(workflow_dir.name.replace("input", "output")) / setting_workflow
    )
    output_path.mkdir(parents=True, exist_ok=True)
    logger.info(f"Setting workflow output path to {output_path}")
    log_file = output_path / "simtools.log"
    configurations = application_config.get("APPLICATIONS")
    for config in configurations:
        for key, value in config.get("CONFIGURATION", {}).items():
            if isinstance(value, str):
                config["CONFIGURATION"][key] = value.replace(place_holder, setting_workflow)
            if isinstance(value, list):
                config["CONFIGURATION"][key] = [
                    item.replace(place_holder, setting_workflow) for item in value
                ]
        config["CONFIGURATION"]["USE_PLAIN_OUTPUT_PATH"] = True
        config["CONFIGURATION"]["OUTPUT_PATH"] = str(output_path)

    return configurations, log_file


def main():  # noqa: D103
    args_dict, db_config = _parse(
        Path(__file__).stem,
        description="Run simtools applications from configuration file.",
        usage="simtools-run-application --config_file config_file_name",
    )
    logger = logging.getLogger()
    logger.setLevel(gen.get_log_level_from_user(args_dict["log_level"]))

    configurations, log_file = read_application_configuration(
        args_dict["configuration_file"], logger
    )

    with log_file.open("w", encoding="utf-8") as file:
        file.write("Running simtools applications\n")
        file.write(dependencies.get_version_string(db_config))
        for config in configurations:
            logger.info(f"Running application: {config.get('APPLICATION')}")
            config = gen.change_dict_keys_case(config, False)
            stdout, stderr = run_application(
                config.get("APPLICATION"), config.get("CONFIGURATION"), logger
            )
            file.write("=" * 80 + "\n")
            file.write(f"Application: {config.get('APPLICATION')}\n")
            file.write("STDOUT:\n" + stdout)
            file.write("STDERR:\n" + stderr)


if __name__ == "__main__":
    main()



#!/usr/bin/python3

r"""
Generate a run script and submit file for HT Condor job submission of a simulation production.

This tool facilitates the submission of multiple simulations to the HT Condor batch system,
enabling:

- Execution of simulations using the "simtools-simulate-prod" application.
- 'number_of_runs' jobs are submitted to the HT Condor batch system.
- Utilization of an Apptainer image containing the SimPipe simulation software and tools.
- Packaging of data and histogram files, and writing them to a specified directory.

This tool is intended for use in an HT Condor environment. Jobs run in a container universe
using the Apptainer image specified in the command line ('--apptainer_image'). Output is written
to the 'output_path' directory, with 'simtools-output' and 'logs' subdirectories.

Requirements for the 'simtools-simulate-prod-htcondor-generator' application:

- Availability of an Apptainer image 'simtools-prod' (obtainable from the package registry on
  GitHub, e.g., via 'apptainer pull --force docker://ghcr.io/gammasim/simtools-prod:latest').
- Environment parameters required to run CORSIKA and sim_telarray, as well as DB access
  credentials.  These should be listed similarly to a '.env' file and copied to
  'output_path/env.txt'.  Ensure that the path to the simulation software is correctly set to
  'SIMTOOLS_SIMTEL_PATH=/workdir/sim_telarray'.


Command line arguments
----------------------
output_path (str, required)
    Directory where the output and the simulation data files will be written.
apptainer_image (str, optional)
    Apptainer image to use for the simulation (full path).
priority (int, optional)
    Job priority (default: 1).

(all other command line arguments are identical to those of :ref:`simulate_prod`).

"""

import logging

import simtools.utils.general as gen
from simtools.configuration import configurator
from simtools.job_execution import htcondor_script_generator


def _parse(description=None):
    """
    Parse command line configuration.

    Parameters
    ----------
    description: str
        Application description.

    Returns
    -------
    CommandLineParser
        Command line parser object.

    """
    config = configurator.Configurator(description=description)
    config.parser.add_argument(
        "--apptainer_image",
        help="Apptainer image to use for the simulation (full path).",
        type=str,
        required=False,
    )
    config.parser.add_argument(
        "--priority",
        help="Job priority.",
        type=int,
        required=False,
        default=1,
    )
    return config.initialize(
        db_config=False,
        job_submission=False,
        simulation_model=["site", "layout", "telescope", "model_version"],
        simulation_configuration={"software": None, "corsika_configuration": ["all"]},
    )


def main():  # noqa: D103
    args_dict, _ = _parse(description="Prepare simulations production for HT Condor job submission")

    logger = logging.getLogger()
    logger.setLevel(gen.get_log_level_from_user(args_dict["log_level"]))

    htcondor_script_generator.generate_submission_script(args_dict)


if __name__ == "__main__":
    main()



"""Interface to workload managers like gridengine or HTCondor."""

import logging
import subprocess
from pathlib import Path

import simtools.utils.general as gen

__all__ = ["JobExecutionError", "JobManager"]


class JobExecutionError(Exception):
    """Job execution error."""


class JobManager:
    """
    Interface to workload managers like gridengine or HTCondor.

    Expects that jobs are described by shell scripts.

    Parameters
    ----------
    submit_engine : str
        Job submission system. Default is local.
    test : bool
        Testing mode without sub submission.
    """

    engines = {
        "gridengine": "qsub",
        "htcondor": "condor_submit",
        "local": "",
        "test_wms": "test_wms",  # used for testing only
    }

    def __init__(self, submit_engine=None, submit_options=None, test=False):
        """Initialize JobManager."""
        self._logger = logging.getLogger(__name__)
        self.submit_engine = submit_engine
        self.submit_options = submit_options
        self.test = test
        self.run_script = None
        self.run_out_file = None

        self.check_submission_system()

    @property
    def submit_engine(self):
        """Get the submit command."""
        return self._submit_engine

    @submit_engine.setter
    def submit_engine(self, value):
        """
        Set the submit command.

        Parameters
        ----------
        value : str
            Name of submit engine.

        Raises
        ------
        ValueError
            if invalid submit engine.
        """
        self._submit_engine = value or "local"
        if self._submit_engine not in self.engines:
            raise ValueError(f"Invalid submit command: {self._submit_engine}")

    def check_submission_system(self):
        """
        Check that the requested workload manager exist on the system.

        Raises
        ------
        JobExecutionError
            if workflow manager is not found.
        """
        try:
            if self.submit_engine in (None, "local") or gen.program_is_executable(
                self.engines[self.submit_engine]
            ):
                return
        except KeyError:
            pass
        raise JobExecutionError(f"Submit engine {self.submit_engine} not found")

    def submit(self, run_script=None, run_out_file=None, log_file=None):
        """
        Submit a job described by a shell script.

        Parameters
        ----------
        run_script: str
            Shell script describing the job to be submitted.
        run_out_file: str or Path
            Redirect output/error/job stream to this file (out,err,job suffix).
        log_file: str or Path
            The log file of the actual simulator (CORSIKA or sim_telarray).
            Provided in order to print the log excerpt in case of run time error.
        """
        self.run_script = str(run_script)
        run_out_file = Path(run_out_file)
        self.run_out_file = str(run_out_file.parent.joinpath(run_out_file.stem))

        self._logger.info(f"Submitting script {self.run_script}")
        self._logger.info(f"Job output stream {self.run_out_file + '.out'}")
        self._logger.info(f"Job error stream {self.run_out_file + '.err'}")
        self._logger.info(f"Job log stream {self.run_out_file + '.job'}")

        submit_result = 0
        if self.submit_engine == "local":
            submit_result = self._submit_local(log_file)
        else:
            submit_result = getattr(self, f"_submit_{self.submit_engine}")()

        if submit_result != 0:
            raise JobExecutionError(f"Job submission failed with return code {submit_result}")

    def _submit_local(self, log_file):
        """
        Run a job script on the command line (no submission to a workload manager).

        Parameters
        ----------
        log_file: str or Path
            The log file of the actual simulator (CORSIKA or sim_telarray).
            Provided in order to print the log excerpt in case of run time error.

        Returns
        -------
        int
            Return code of the executed script
        """
        self._logger.info("Running script locally")

        if self.test:
            self._logger.info("Testing (local)")
            return 0

        result = None
        try:
            with (
                open(f"{self.run_out_file}.out", "w", encoding="utf-8") as stdout,
                open(f"{self.run_out_file}.err", "w", encoding="utf-8") as stderr,
            ):
                result = subprocess.run(
                    f"{self.run_script}",
                    shell=True,
                    check=True,
                    text=True,
                    stdout=stdout,
                    stderr=stderr,
                )
        except subprocess.CalledProcessError as exc:
            self._logger.error(gen.get_log_excerpt(f"{self.run_out_file}.err"))
            if log_file.exists() and gen.get_file_age(log_file) < 5:
                self._logger.error(gen.get_log_excerpt(log_file))
            raise JobExecutionError("See excerpt from log file above\n") from exc

        return result.returncode if result else 0

    def _submit_htcondor(self):
        """Submit a job described by a shell script to HTcondor."""
        _condor_file = self.run_script + ".condor"
        lines = [
            f"Executable = {self.run_script}",
            f"Output = {self.run_out_file}.out",
            f"Error = {self.run_out_file}.err",
            f"Log = {self.run_out_file}.job",
        ]
        if self.submit_options:
            lines.extend(option.lstrip() for option in self.submit_options.split(","))
        lines.append("queue 1")
        try:
            with open(_condor_file, "w", encoding="utf-8") as file:
                file.write("\n".join(lines) + "\n")
        except FileNotFoundError as exc:
            self._logger.error(f"Failed creating condor submission file {_condor_file}")
            raise JobExecutionError from exc

        return self._execute(self.submit_engine, [self.engines[self.submit_engine], _condor_file])

    def _submit_gridengine(self):
        """Submit a job described by a shell script to gridengine."""
        this_sub_cmd = [
            self.engines[self.submit_engine],
            "-o",
            self.run_out_file + ".out",
            "-e",
            self.run_out_file + ".err",
            self.run_script,
        ]
        return self._execute(self.submit_engine, this_sub_cmd)

    def _execute(self, engine, shell_command):
        """
        Execute a shell command using a specific engine.

        Parameters
        ----------
        engine : str
            Engine to use.
        shell_command : list
            List of shell command plus arguments.
        """
        self._logger.info(f"Submitting script to {engine}")
        self._logger.debug(shell_command)
        result = None
        if not self.test:
            result = subprocess.run(shell_command, shell=True, check=True)
        else:
            self._logger.info(f"Testing ({engine}: {shell_command})")

        return result.returncode if result else 0



#!/usr/bin/python3

r"""
    Derives the mirror alignment parameters using cumulative PSF measurement.

    This includes parameters mirror_reflection_random_angle, \
    mirror_align_random_horizontal and mirror_align_random_vertical.

    The telescope zenith angle and the source distance can be set by command line arguments.

    The measured cumulative PSF should be provided by using the command line argument data. \
    A file name is expected, in which the file should contain 3 columns: radial distance in mm, \
    differential value of photon intensity and its integral value.

    The derivation is performed through a random search. A number of random combination of the \
    parameters are tested and the best ones are selected based on the minimum value of \
    the Root Mean Squared Deviation between data and simulations. The range in which the \
    parameter are drawn uniformly are defined based on the previous value on the telescope model.

    The assumption are:

    a) mirror_align_random_horizontal and mirror_align_random_vertical are the same.

    b) mirror_align_random_horizontal/vertical have no dependence on the zenith angle.

    One example of the plot generated by this applications are shown below.

    .. _derive_psf_parameters_plot:
    .. image::  images/derive_psf_parameters.png
      :width: 49 %

    Command line arguments
    ----------------------
    site (str, required)
        North or South.
    telescope (str, required)
        Telescope model name (e.g. LST-1, SST-D, ...).
    model_version (str, optional)
        Model version.
    src_distance (float, optional)
        Source distance in km.
    zenith (float, optional)
        Zenith angle in deg.
    data (str, optional)
        Name of the data file with the measured cumulative PSF.
    plot_all (activation mode, optional)
        If activated, plots will be generated for all values tested during tuning.
    fixed (activation mode, optional)
        Keep the first entry of mirror_reflection_random_angle fixed.
    test (activation mode, optional)
        If activated, application will be faster by simulating fewer photons.

    Example
    -------
    LSTN-01 5.0.0

    Runtime < 3 min.

    Get PSF data from the DB:

    .. code-block:: console

        simtools-db-get-file-from-db --file_name PSFcurve_data_v2.txt

    Run the application:

    .. code-block:: console

        simtools-derive-psf-parameters --site North --telescope LSTN-01 \\
            --model_version 6.0.0 --data PSFcurve_data_v2.txt --plot_all --test

    The output is saved in simtools-output/derive_psf_parameters.

    Expected final print-out message:

    .. code-block:: console

        Best parameters:
        mirror_reflection_random_angle = [0.006, 0.133, 0.005]
        mirror_align_random_horizontal = [0.005, 28.0, 0.0, 0.0]
        mirror_align_random_vertical = [0.005, 28.0, 0.0, 0.0]

"""

import logging
from collections import OrderedDict

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.backends.backend_pdf import PdfPages

import simtools.utils.general as gen
from simtools.configuration import configurator
from simtools.io_operations import io_handler
from simtools.model.telescope_model import TelescopeModel
from simtools.ray_tracing.ray_tracing import RayTracing
from simtools.visualization import visualize


def load_data(data_file):
    """
    Load data from a text file containing cumulative PSF measurements.

    Parameters
    ----------
    data_file : str
        Name of the data file with the measured cumulative PSF.

    Returns
    -------
    numpy.ndarray
        Loaded and processed data from the file.
    """
    radius_cm = "Radius [cm]"
    cumulative_psf = "Cumulative PSF"
    d_type = {"names": (radius_cm, cumulative_psf), "formats": ("f8", "f8")}
    data = np.loadtxt(data_file, dtype=d_type, usecols=(0, 2))
    data[radius_cm] *= 0.1
    data[cumulative_psf] /= np.max(np.abs(data[cumulative_psf]))
    return data


def _parse():
    config = configurator.Configurator(
        description=(
            "Derive mirror_reflection_random_angle, mirror_align_random_horizontal "
            "and mirror_align_random_vertical using cumulative PSF measurement."
        )
    )
    config.parser.add_argument(
        "--src_distance",
        help="Source distance in km",
        type=float,
        default=10,
    )
    config.parser.add_argument("--zenith", help="Zenith angle in deg", type=float, default=20)
    config.parser.add_argument(
        "--data", help="Data file name with the measured PSF vs radius [cm]", type=str
    )
    config.parser.add_argument(
        "--plot_all",
        help=(
            "On: plot cumulative PSF for all tested combinations, "
            "Off: plot it only for the best set of values"
        ),
        action="store_true",
    )
    config.parser.add_argument(
        "--fixed",
        help=("Keep the first entry of mirror_reflection_random_angle fixed."),
        action="store_true",
    )
    return config.initialize(db_config=True, simulation_model=["telescope", "model_version"])


def add_parameters(
    all_parameters,
    mirror_reflection,
    mirror_align,
    mirror_reflection_fraction=0.15,
    mirror_reflection_2=0.035,
):
    """
    Transform and add parameters to the all_parameters list.

    Parameters
    ----------
    mirror_reflection : float
        The random angle of mirror reflection.

    mirror_align : float
        The random angle for mirror alignment (both horizontal and vertical).

    mirror_reflection_fraction : float, optional
        The fraction of the mirror reflection. Default is 0.15.

    mirror_reflection_2 : float, optional
        A secondary random angle for mirror reflection. Default is 0.035.

    Returns
    -------
    None
        Updates the all_parameters list in place.
    """
    # If we want to start from values different than the ones currently in the model:
    # align = 0.0046
    # pars_to_change = {
    #     'mirror_reflection_random_angle': '0.0075 0.125 0.0037',
    #     'mirror_align_random_horizontal': f'{align} 28 0 0',
    #     'mirror_align_random_vertical': f'{align} 28 0 0',
    # }
    # tel_model.change_multiple_parameters(**pars_to_change)
    pars = {
        "mirror_reflection_random_angle": [
            mirror_reflection,
            mirror_reflection_fraction,
            mirror_reflection_2,
        ],
        "mirror_align_random_horizontal": [mirror_align, 28.0, 0.0, 0.0],
        "mirror_align_random_vertical": [mirror_align, 28.0, 0.0, 0.0],
    }
    all_parameters.append(pars)


def get_previous_values(tel_model, logger):
    """
    Retrieve previous parameter values from the telescope model.

    Parameters
    ----------
    tel_model : TelescopeModel
        Telescope model object.
    logger : logging.Logger
        Logger object for logging messages.

    Returns
    -------
    tuple
        Tuple containing the previous values of mirror_reflection_random_angle (first entry),
        mirror_reflection_fraction, second entry), mirror_reflection_random_angle (third entry),
        and mirror_align_random_horizontal/vertical.
    """
    split_par = tel_model.get_parameter_value("mirror_reflection_random_angle")
    mrra_0, mfr_0, mrra2_0 = split_par[0], split_par[1], split_par[2]
    mar_0 = tel_model.get_parameter_value("mirror_align_random_horizontal")[0]

    logger.debug(
        "Previous parameter values:\n"
        f"MRRA = {mrra_0!s}\n"
        f"MRF = {mfr_0!s}\n"
        f"MRRA2 = {mrra2_0!s}\n"
        f"MAR = {mar_0!s}\n"
    )

    return mrra_0, mfr_0, mrra2_0, mar_0


def generate_random_parameters(
    all_parameters, n_runs, args_dict, mrra_0, mfr_0, mrra2_0, mar_0, logger
):
    """
    Generate random parameters for tuning.

    Parameters
    ----------
    all_parameters : list
        List to store all parameter sets.
    n_runs : int
        Number of random parameter combinations to test.
    args_dict : dict
        Dictionary containing parsed command-line arguments.
    mrra_0 : float
        Initial value of mirror_reflection_random_angle.
    mfr_0 : float
        Initial value of mirror_reflection_fraction.
    mrra2_0 : float
        Initial value of the second mirror_reflection_random_angle.
    mar_0 : float
        Initial value of mirror_align_random_horizontal/vertical.
    logger : logging.Logger
        Logger object for logging messages.
    """
    # Range around the previous values are hardcoded
    # Number of runs is hardcoded
    if args_dict["fixed"]:
        logger.debug("fixed=True - First entry of mirror_reflection_random_angle is kept fixed.")

    for _ in range(n_runs):
        mrra_range = 0.004 if not args_dict["fixed"] else 0
        mrf_range = 0.1
        mrra2_range = 0.03
        mar_range = 0.005
        rng = np.random.default_rng()
        mrra = rng.uniform(max(mrra_0 - mrra_range, 0), mrra_0 + mrra_range)
        mrf = rng.uniform(max(mfr_0 - mrf_range, 0), mfr_0 + mrf_range)
        mrra2 = rng.uniform(max(mrra2_0 - mrra2_range, 0), mrra2_0 + mrra2_range)
        mar = rng.uniform(max(mar_0 - mar_range, 0), mar_0 + mar_range)
        add_parameters(all_parameters, mrra, mar, mrf, mrra2)


def load_and_process_data(args_dict):
    """
    Load and process data if specified in the command-line arguments.

    Returns
    -------
    - data_to_plot: OrderedDict containing loaded and processed data.
    - radius: Radius data from loaded data (if available).
    """
    data_to_plot = OrderedDict()
    radius = None

    if args_dict["data"] is not None:
        data_file = gen.find_file(args_dict["data"], args_dict["model_path"])
        data_to_plot["measured"] = load_data(data_file)
        radius = data_to_plot["measured"]["Radius [cm]"]

    return data_to_plot, radius


def calculate_rmsd(data, sim):
    """Calculate Root Mean Squared Deviation to be used as metric to find the best parameters."""
    return np.sqrt(np.mean((data - sim) ** 2))


def run_pars(tel_model, args_dict, pars, data_to_plot, radius, pdf_pages):
    """
    Run the tuning for one set of parameters, add a plot to the pdfPages and return RMSD and D80.

    Plotting is optional (if plot=True).
    """
    cumulative_psf = "Cumulative PSF"

    if pars is not None:
        tel_model.change_multiple_parameters(**pars)
    else:
        raise ValueError("No best parameters found")

    ray = RayTracing(
        telescope_model=tel_model,
        simtel_path=args_dict["simtel_path"],
        zenith_angle=args_dict["zenith"] * u.deg,
        source_distance=args_dict["src_distance"] * u.km,
        off_axis_angle=[0.0] * u.deg,
    )

    ray.simulate(test=args_dict["test"], force=True)
    ray.analyze(force=True, use_rx=False)

    # Plotting cumulative PSF
    im = ray.images()[0]
    d80 = im.get_psf()

    if radius is not None:
        # Simulated cumulative PSF
        data_to_plot["simulated"] = im.get_cumulative_data(radius * u.cm)
    else:
        raise ValueError("Radius data is not available.")

    rmsd = calculate_rmsd(
        data_to_plot["measured"][cumulative_psf], data_to_plot["simulated"][cumulative_psf]
    )

    if args_dict["plot_all"]:
        fig = visualize.plot_1d(
            data_to_plot,
            plot_difference=True,
            no_markers=True,
        )
        ax = fig.get_axes()[0]
        ax.set_ylim(0, 1.05)
        ax.set_title(
            f"refl_rnd={pars['mirror_reflection_random_angle']}, "
            f"align_rnd={pars['mirror_align_random_vertical']}"
        )

        ax.text(
            0.8,
            0.3,
            f"D80 = {d80:.3f} cm\nRMSD = {rmsd:.4f}",
            verticalalignment="center",
            horizontalalignment="center",
            transform=ax.transAxes,
        )
        plt.tight_layout()
        pdf_pages.savefig(fig)
        plt.clf()

    return d80, rmsd


def find_best_parameters(all_parameters, tel_model, args_dict, data_to_plot, radius, pdf_pages):
    """
    Find the best parameters from all parameter sets.

    Returns
    -------
    - Tuple of best parameters and their D80 value.
    """
    min_rmsd = 100
    best_pars = None

    for pars in all_parameters:
        _, rmsd = run_pars(tel_model, args_dict, pars, data_to_plot, radius, pdf_pages)
        if rmsd < min_rmsd:
            min_rmsd = rmsd
            best_pars = pars

    return best_pars, min_rmsd


def main():  # noqa: D103
    args_dict, db_config = _parse()

    label = "tune_psf"
    logger = logging.getLogger()
    logger.setLevel(gen.get_log_level_from_user(args_dict["log_level"]))

    # Output directory to save files related directly to this app
    _io_handler = io_handler.IOHandler()
    output_dir = _io_handler.get_output_directory(label, sub_dir="application-plots")
    tel_model = TelescopeModel(
        site=args_dict["site"],
        telescope_name=args_dict["telescope"],
        mongo_db_config=db_config,
        model_version=args_dict["model_version"],
        label=label,
    )

    all_parameters = []
    mrra_0, mfr_0, mrra2_0, mar_0 = get_previous_values(tel_model, logger)

    n_runs = 5 if args_dict["test"] else 50
    generate_random_parameters(
        all_parameters, n_runs, args_dict, mrra_0, mfr_0, mrra2_0, mar_0, logger
    )

    data_to_plot, radius = load_and_process_data(args_dict)

    # Preparing figure name
    plot_file_name = "_".join((label, tel_model.name + ".pdf"))
    plot_file = output_dir.joinpath(plot_file_name)
    pdf_pages = PdfPages(plot_file)

    best_pars, _ = find_best_parameters(
        all_parameters, tel_model, args_dict, data_to_plot, radius, pdf_pages
    )

    # Rerunning and plotting the best pars
    run_pars(tel_model, args_dict, best_pars, data_to_plot, radius, pdf_pages)
    plt.close()
    pdf_pages.close()

    # Printing the results
    print("Best parameters:")
    for par, value in best_pars.items():
        print(f"{par} = {value}")


if __name__ == "__main__":
    main()



"""Simulation using the light emission package for calibration."""

import logging
import stat
from pathlib import Path

import astropy.units as u
import numpy as np

from simtools.io_operations import io_handler
from simtools.runners.simtel_runner import SimtelRunner
from simtools.utils.general import clear_default_sim_telarray_cfg_directories

__all__ = ["SimulatorLightEmission"]


class SimulatorLightEmission(SimtelRunner):
    """
    Interface with sim_telarray to perform light emission package simulations.

    The light emission package is used to simulate a artificial light source, used for calibration.

    The angle and pointing vector calculations use the convention north (x) towards
    east (-y).

    Parameters
    ----------
    telescope_model:
        Instance of the TelescopeModel class.
    calibration_model:
        CalibrationModel instance to define calibration device.
    site_model:
        SiteModel instance to define the site specific parameters.
    default_le_config: dict
        defines parameters for running the sim_telarray light emission application.
    le_application: str
        Name of the application. Default sim_telarray application running
        the sim_telarray LightEmission package is xyzls.
    simtel_path: str or Path
        Location of sim_telarray installation.
    light_source_type: str
        The light source type.
    label: str
        Label for output directory.
    config_data: dict
        Dict containing the configurable parameters.
    config_file: str or Path
        Path of the yaml file containing the configurable parameters.
    test: bool
        Test flag.
    """

    def __init__(
        self,
        telescope_model,
        calibration_model,
        site_model,
        default_le_config,
        le_application,
        simtel_path,
        light_source_type,
        label=None,
        test=False,
    ):
        """Initialize SimtelRunner."""
        super().__init__(label=label, simtel_path=simtel_path)

        self._logger = logging.getLogger(__name__)
        self._logger.debug("Init SimtelRunnerLightEmission")
        self._simtel_path = simtel_path

        self._telescope_model = telescope_model

        self.label = label if label is not None else self._telescope_model.label

        self._calibration_model = calibration_model
        self._site_model = site_model
        self.io_handler = io_handler.IOHandler()
        self.output_directory = self.io_handler.get_output_directory(self.label)

        # LightEmission - default parameters
        self._rep_number = 0
        self.runs = 1
        self.photons_per_run = (
            self._calibration_model.get_parameter_value("photons_per_run") if not test else 1e7
        )

        self.le_application = le_application
        self.default_le_config = default_le_config
        self.distance = self.telescope_calibration_device_distance()
        self.light_source_type = light_source_type
        self._telescope_model.export_config_file()
        self.test = test

    @staticmethod
    def light_emission_default_configuration():
        """
        Get default light emission configuration.

        Returns
        -------
        dict
            Default configuration light emission.

        """
        return {
            "zenith_angle": {
                "len": 1,
                "unit": u.Unit("deg"),
                "default": 0.0 * u.deg,
                "names": ["zenith", "theta"],
            },
            "azimuth_angle": {
                "len": 1,
                "unit": u.Unit("deg"),
                "default": 0.0 * u.deg,
                "names": ["azimuth", "phi"],
            },
            "source_distance": {
                "len": 1,
                "unit": u.Unit("m"),
                "default": 1000 * u.m,
                "names": ["sourcedist", "srcdist"],
            },
            "off_axis_angle": {
                "len": 1,
                "unit": u.Unit("deg"),
                "default": 0 * u.deg,
                "names": ["off_axis"],
            },
            "fadc_bins": {
                "len": 1,
                "unit": u.dimensionless_unscaled,
                "default": 128,
                "names": ["fadc_bins"],
            },
        }

    def calibration_pointing_direction(self):
        """
        Calculate the pointing of the calibration device towards the telescope.

        Returns
        -------
        list
            The pointing vector from the calibration device to the telescope.
        """
        # use DB coordinates later
        x_cal, y_cal, z_cal = self._calibration_model.get_parameter_value(
            "array_element_position_ground"
        )
        cal_vect = np.array([x_cal, y_cal, z_cal])
        x_tel, y_tel, z_tel = self._telescope_model.get_parameter_value(
            "array_element_position_ground"
        )

        tel_vect = np.array([x_tel, y_tel, z_tel])

        direction_vector = tel_vect - cal_vect
        # pointing vector from calibration device to telescope

        pointing_vector = np.round(direction_vector / np.linalg.norm(direction_vector), 6)

        # Calculate telescope theta and phi angles
        tel_theta = 180 - np.round(
            np.rad2deg(np.arccos(direction_vector[2] / np.linalg.norm(direction_vector))), 6
        )
        tel_phi = 180 - np.round(
            np.rad2deg(np.arctan2(direction_vector[1], direction_vector[0])), 6
        )

        # Calculate laser beam theta and phi angles
        direction_vector_inv = direction_vector * -1
        laser_theta = np.round(
            np.rad2deg(np.arccos(direction_vector_inv[2] / np.linalg.norm(direction_vector_inv))), 6
        )
        laser_phi = np.round(
            np.rad2deg(np.arctan2(direction_vector_inv[1], direction_vector_inv[0])), 6
        )
        return pointing_vector.tolist(), [tel_theta, tel_phi, laser_theta, laser_phi]

    def telescope_calibration_device_distance(self):
        """
        Calculate the distance between the telescope and the calibration device.

        Returns
        -------
        astropy Quantity
            The distance between the telescope and the calibration device.
        """
        if not self.default_le_config:
            x_cal, y_cal, z_cal = self._calibration_model.get_parameter_value(
                "array_element_position_ground"
            )
            x_tel, y_tel, z_tel = self._telescope_model.get_parameter_value(
                "array_element_position_ground"
            )

        else:
            x_tel = self.default_le_config["x_pos"]["default"].to(u.m).value
            y_tel = self.default_le_config["y_pos"]["default"].to(u.m).value
            z_tel = self.default_le_config["z_pos"]["default"].to(u.m).value

            x_cal, y_cal, z_cal = 0, 0, 0

        tel_vect = np.array([x_tel, y_tel, z_tel])
        cal_vect = np.array([x_cal, y_cal, z_cal])
        distance = np.linalg.norm(cal_vect - tel_vect)

        return distance * u.m

    def _make_light_emission_script(self):
        """
        Create the light emission script to run the light emission package.

        Require the specified pre-compiled light emission package application
        in the sim_telarray/LightEmission/ path.

        Returns
        -------
        str
            The commands to run the Light Emission package
        """
        x_cal, y_cal, z_cal = (
            self._calibration_model.get_parameter_value("array_element_position_ground") * u.m
        )
        x_tel, y_tel, z_tel = (
            self._telescope_model.get_parameter_value("array_element_position_ground") * u.m
        )
        _model_directory = self.io_handler.get_output_directory(self.label, "model")
        command = f" rm {self.output_directory}/"
        command += f"{self.le_application[0]}_{self.le_application[1]}.simtel.gz\n"
        command += str(self._simtel_path.joinpath("sim_telarray/LightEmission/"))
        command += f"/{self.le_application[0]}"

        if self.light_source_type == "led":
            if self.le_application[1] == "variable":
                command += f" -x {self.default_le_config['x_pos']['default'].to(u.cm).value}"
                command += f" -y {self.default_le_config['y_pos']['default'].to(u.cm).value}"
                command += f" -z {self.default_le_config['z_pos']['default'].to(u.cm).value}"
                command += (
                    f" -d {','.join(map(str, self.default_le_config['direction']['default']))}"
                )
                command += f" -n {self.photons_per_run}"

            elif self.le_application[1] == "layout":
                x_origin = x_cal - x_tel
                y_origin = y_cal - y_tel
                z_origin = z_cal - z_tel
                # light_source coordinates relative to telescope
                command += f" -x {x_origin.to(u.cm).value}"
                command += f" -y {y_origin.to(u.cm).value}"
                command += f" -z {z_origin.to(u.cm).value}"
                pointing_vector = self.calibration_pointing_direction()[0]
                command += f" -d {','.join(map(str, pointing_vector))}"

                command += f" -n {self.photons_per_run}"

                # same wavelength as for laser
                command += f" -s {self._calibration_model.get_parameter_value('laser_wavelength')}"

                # pulse
                command += (
                    f" -p Gauss:{self._calibration_model.get_parameter_value('led_pulse_sigtime')}"
                )
                command += " -a isotropic"  # angular distribution

            command += f" -A {_model_directory}/"
            command += f"{self._telescope_model.get_parameter_value('atmospheric_profile')}"

        elif self.light_source_type == "laser":
            command += " --events 1"
            command += " --bunches 2500000"
            command += " --step 0.1"
            command += " --bunchsize 1"
            command += (
                f" --spectrum {self._calibration_model.get_parameter_value('laser_wavelength')}"
            )
            command += " --lightpulse Gauss:"
            command += f"{self._calibration_model.get_parameter_value('laser_pulse_sigtime')}"
            x_origin = x_cal - x_tel
            y_origin = y_cal - y_tel
            z_origin = z_cal - z_tel
            _, angles = self.calibration_pointing_direction()
            angle_theta = angles[0]
            angle_phi = angles[1]
            positions = x_origin.to(u.cm).value, y_origin.to(u.cm).value, z_origin.to(u.cm).value
            command += f" --laser-position '{positions[0]},{positions[1]},{positions[2]}'"

            command += f" --telescope-theta {angle_theta}"
            command += f" --telescope-phi {angle_phi}"
            command += f" --laser-theta {90 - angles[2]}"
            command += f" --laser-phi {angles[3]}"  # convention north (x) towards east (-y)
            command += f" --atmosphere {_model_directory}/"
            command += f"{self._telescope_model.get_parameter_value('atmospheric_profile')}"
        command += f" -o {self.output_directory}/{self.le_application[0]}.iact.gz"
        command += "\n"

        return command

    def _make_simtel_script(self):
        """
        Return the command to run simtel_array using the output from the previous step.

        Returns
        -------
        str
            The command to run simtel_array
        """
        # LightEmission
        _, angles = self.calibration_pointing_direction()

        command = f"{self._simtel_path.joinpath('sim_telarray/bin/sim_telarray/')}"
        command += " -I"
        command += f" -I{self._telescope_model.config_file_directory}"
        command += f" -c {self._telescope_model.get_config_file(no_export=True)}"
        if not self.test:
            self._remove_line_from_config(
                self._telescope_model.get_config_file(no_export=True), "array_triggers"
            )
            self._remove_line_from_config(
                self._telescope_model.get_config_file(no_export=True), "axes_offsets"
            )

        command += " -DNUM_TELESCOPES=1"

        command += super().get_config_option(
            "altitude", self._site_model.get_parameter_value("corsika_observation_level")
        )
        command += super().get_config_option(
            "atmospheric_transmission",
            self._telescope_model.get_parameter_value("atmospheric_transmission"),
        )
        command += super().get_config_option("TRIGGER_TELESCOPES", "1")

        command += super().get_config_option("TELTRIG_MIN_SIGSUM", "2")
        command += super().get_config_option("PULSE_ANALYSIS", "-30")
        command += super().get_config_option("MAXIMUM_TELESCOPES", 1)

        if self.le_application[1] == "variable":
            command += super().get_config_option("telescope_theta", 0)
            command += super().get_config_option("telescope_phi", 0)
        else:
            command += super().get_config_option("telescope_theta", f"{angles[0]}")
            command += super().get_config_option("telescope_phi", f"{angles[1]}")

        command += super().get_config_option("power_law", "2.68")
        command += super().get_config_option(
            "input_file", f"{self.output_directory}/{self.le_application[0]}.iact.gz"
        )
        command += super().get_config_option(
            "output_file",
            f"{self.output_directory}/{self.le_application[0]}_{self.le_application[1]}.simtel.gz",
        )
        command += super().get_config_option(
            "histogram_file",
            f"{self.output_directory}/"
            f"{self.le_application[0]}_{self.le_application[1]}.ctsim.hdata\n",
        )

        # Remove the default sim_telarray configuration directories
        return clear_default_sim_telarray_cfg_directories(command)

    def _remove_line_from_config(self, file_path, line_prefix):
        """
        Remove lines starting with a specific prefix from the config.

        Parameters
        ----------
        file_path : Path
            The path to the configuration file.
        line_prefix : str
            The prefix of lines to be removed.
        """
        file_path = Path(file_path)
        with file_path.open("r", encoding="utf-8") as file:
            lines = file.readlines()

        with file_path.open("w", encoding="utf-8") as file:
            for line in lines:
                if not line.startswith(line_prefix):
                    file.write(line)

    def _create_postscript(self, **kwargs):
        """
        Write out post-script file using read_cta in hessioxxx/bin/read_cta.

        parts from the documentation
        -r level        (Use 10/5 tail-cut image cleaning and redo reconstruction.)
                level >= 1: show parameters from sim_hessarray.
                level >= 2: redo shower reconstruction
                level >= 3: redo image cleaning (and shower reconstruction
                            with new image parameters)
                level >= 4: redo amplitude summation
                level >= 5: PostScript file includes original and
                            new shower reconstruction.
        --integration-window w,o[,ps] *(Set integration window width and offset.)
            For some integration schemes there is a pulse shaping option.


        Returns
        -------
        str
            Command to create the postscript file
        """
        postscript_dir = self.output_directory.joinpath("postscripts")
        postscript_dir.mkdir(parents=True, exist_ok=True)

        command = str(self._simtel_path.joinpath("hessioxxx/bin/read_cta"))
        command += " --min-tel 1 --min-trg-tel 1"
        command += " -q --integration-scheme 4"
        command += " --integration-window "
        command += f"{kwargs['integration_window'][0]},{kwargs['integration_window'][1]}"
        command += f" -r {kwargs['level']}"
        command += " --plot-with-sum-only"
        command += " --plot-with-pixel-amp --plot-with-pixel-id"
        command += (
            f" -p {postscript_dir}/"
            f"{self.le_application[0]}_{self.le_application[1]}_"
            f"d_{int(self.distance.to(u.m).value)}.ps"
        )
        command += (
            f" {self.output_directory}/"
            f"{self.le_application[0]}_{self.le_application[1]}.simtel.gz\n"
        )
        # ps2pdf required, now only store postscripts
        # command += f"ps2pdf {self.output_directory}/{self.le_application}.ps
        #  {self.output_directory}/{self.le_application}.pdf"
        return command

    def prepare_script(self, generate_postscript=False, **kwargs):
        """
        Build and return bash run script containing the light-emission command.

        Parameters
        ----------
        plot: bool
            If output should be plotted.

        generate_postscript: bool
            If postscript should be generated with read_cta.

        Returns
        -------
        Path
            Full path of the run script.
        """
        self._logger.debug("Creating run bash script")

        _script_dir = self.output_directory.joinpath("scripts")
        _script_dir.mkdir(parents=True, exist_ok=True)
        _script_file = _script_dir.joinpath(f"{self.le_application[0]}-lightemission.sh")
        self._logger.debug(f"Run bash script - {_script_file}")

        command_le = self._make_light_emission_script()
        command_simtel = self._make_simtel_script()

        with _script_file.open("w", encoding="utf-8") as file:
            file.write("#!/usr/bin/env bash\n\n")

            file.write(f"{command_le}\n\n")
            file.write(f"{command_simtel}\n\n")

            if generate_postscript:
                self._logger.debug("Write out postscript file")
                command_plot = self._create_postscript(**kwargs)
                file.write("# Generate postscript\n\n")
                file.write(f"{command_plot}\n\n")
                file.write("# End\n\n")

        _script_file.chmod(_script_file.stat().st_mode | stat.S_IXUSR | stat.S_IXGRP)
        return _script_file
